from collections import (
    deque,
    OrderedDict
)
from decimal import Decimal
from typing import (
    Dict,
    List,
    Tuple
)

import pandas as pd

from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.limit_order cimport LimitOrder
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.market_order import MarketOrder
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple

NaN = float("nan")

cdef class OrderTracker(TimeIterator):
    # ETH confirmation requirement of Binance has shortened to 12 blocks as of 7/15/2019.
    # 12 * 15 / 60 = 3 minutes
    SHADOW_MAKER_ORDER_KEEP_ALIVE_DURATION = 60.0 * 3

    CANCEL_EXPIRY_DURATION = 60.0

    def __init__(self):
        super().__init__()
        self._tracked_limit_orders = {}
        self._tracked_market_orders = {}
        self._order_id_to_market_pair = {}
        self._shadow_tracked_limit_orders = {}
        self._shadow_order_id_to_market_pair = {}
        self._shadow_gc_requests = deque()
        self._in_flight_pending_created = set()
        self._in_flight_cancels = OrderedDict()

    @property
    def active_limit_orders(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        limit_orders = []
        for market_pair, orders_map in self._tracked_limit_orders.items():
            for limit_order in orders_map.values():
                if self.c_has_in_flight_cancel(limit_order.client_order_id):
                    continue
                limit_orders.append((market_pair.market, limit_order))
        return limit_orders

    @property
    def shadow_limit_orders(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        limit_orders = []
        for market_pair, orders_map in self._shadow_tracked_limit_orders.items():
            for limit_order in orders_map.values():
                if self.c_has_in_flight_cancel(limit_order.client_order_id):
                    continue
                limit_orders.append((market_pair.market, limit_order))
        return limit_orders

    @property
    def market_pair_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
        market_pair_to_orders = {}
        market_pairs = self._tracked_limit_orders.keys()
        for market_pair in market_pairs:
            limit_orders = []
            for limit_order in self._tracked_limit_orders[market_pair].values():
                if self.c_has_in_flight_cancel(limit_order.client_order_id):
                    continue
                limit_orders.append(limit_order)
            market_pair_to_orders[market_pair] = limit_orders
        return market_pair_to_orders

    @property
    def active_bids(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        return [(market, limit_order) for market, limit_order in self.active_limit_orders if limit_order.is_buy]

    @property
    def active_asks(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        return [(market, limit_order) for market, limit_order in self.active_limit_orders if not limit_order.is_buy]

    @property
    def tracked_limit_orders(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        return [(market_trading_pair_tuple[0], order) for market_trading_pair_tuple, order_map in
                self._tracked_limit_orders.items()
                for order in order_map.values()]

    @property
    def tracked_limit_orders_map(self) -> Dict[ConnectorBase, Dict[str, LimitOrder]]:
        return self._tracked_limit_orders

    @property
    def tracked_limit_orders_data_frame(self) -> pd.DataFrame:
        limit_orders = [
            [market_trading_pair_tuple.market.display_name, market_trading_pair_tuple.trading_pair, order_id,
             order.quantity,
             pd.Timestamp(order.creation_timestamp / 1e6, unit='s', tz='UTC').strftime('%Y-%m-%d %H:%M:%S')
             ]
            for market_trading_pair_tuple, order_map in self._tracked_limit_orders.items()
            for order_id, order in order_map.items()]

        return pd.DataFrame(data=limit_orders, columns=["market", "trading_pair", "order_id", "quantity", "timestamp"])

    @property
    def tracked_market_orders(self) -> List[Tuple[ConnectorBase, MarketOrder]]:
        return [(market_trading_pair_tuple[0], order) for market_trading_pair_tuple, order_map
                in self._tracked_market_orders.items() for order in order_map.values()]

    @property
    def tracked_market_orders_data_frame(self) -> List[pd.DataFrame]:
        market_orders = [[market_trading_pair_tuple.market.display_name, market_trading_pair_tuple.trading_pair,
                          order_id, order.amount,
                          pd.Timestamp(order.timestamp, unit='s', tz='UTC').strftime('%Y-%m-%d %H:%M:%S')]
                         for market_trading_pair_tuple, order_map in self._tracked_market_orders.items()
                         for order_id, order in order_map.items()]

        return pd.DataFrame(data=market_orders, columns=["market", "trading_pair", "order_id", "quantity", "timestamp"])

    @property
    def in_flight_cancels(self) -> Dict[str, float]:
        return self._in_flight_cancels

    @property
    def in_flight_pending_created(self) -> Dict[str, float]:
        return self._in_flight_pending_created

    cdef c_tick(self, double timestamp):
        TimeIterator.c_tick(self, timestamp)
        self.c_check_and_cleanup_shadow_records()

    cdef dict c_get_limit_orders(self):
        return self._tracked_limit_orders

    def get_limit_orders(self):
        return self.c_get_limit_orders()

    cdef dict c_get_market_orders(self):
        return self._tracked_market_orders

    def get_market_orders(self):
        return self.c_get_market_orders()

    cdef dict c_get_shadow_limit_orders(self):
        return self._shadow_tracked_limit_orders

    def get_shadow_limit_orders(self) -> Dict[MarketTradingPairTuple, Dict[str, LimitOrder]]:
        return self.c_get_shadow_limit_orders()

    cdef bint c_has_in_flight_cancel(self, str order_id):
        return self._in_flight_cancels.get(order_id, NaN) + self.CANCEL_EXPIRY_DURATION > self._current_timestamp

    def has_in_flight_cancel(self, order_id: str):
        return self.c_has_in_flight_cancel(order_id)

    cdef bint c_check_and_track_cancel(self, str order_id):
        """
        :param order_id: the order id to be canceled
        :return: True if there's no existing in flight cancel for the order id, False otherwise.
        """
        cdef:
            list keys_to_delete = []

        if order_id in self._in_flight_pending_created:  # Checks if a Buy/SellOrderCreatedEvent has been received
            return False

        # Maintain the cancel expiry time invariant.
        for k, cancel_timestamp in self._in_flight_cancels.items():
            if cancel_timestamp < self._current_timestamp - self.CANCEL_EXPIRY_DURATION:
                keys_to_delete.append(k)
        for k in keys_to_delete:
            del self._in_flight_cancels[k]

        if order_id in self.in_flight_cancels:
            return False

        # Track the cancel.
        self._in_flight_cancels[order_id] = self._current_timestamp
        return True

    def check_and_track_cancel(self, order_id: str) -> bool:
        return self.c_check_and_track_cancel(order_id)

    cdef object c_get_market_pair_from_order_id(self, str order_id):
        return self._order_id_to_market_pair.get(order_id)

    def get_market_pair_from_order_id(self, order_id: str):
        return self.c_get_market_pair_from_order_id(order_id)

    cdef object c_get_shadow_market_pair_from_order_id(self, str order_id):
        return self._shadow_order_id_to_market_pair.get(order_id)

    def get_shadow_market_pair_from_order_id(self, order_id: str) -> MarketTradingPairTuple:
        return self.c_get_shadow_market_pair_from_order_id(order_id)

    cdef LimitOrder c_get_limit_order(self, object market_pair, str order_id):
        return self._tracked_limit_orders.get(market_pair, {}).get(order_id)

    def get_limit_order(self, market_pair, order_id: str) -> LimitOrder:
        return self.c_get_limit_order(market_pair, order_id)

    cdef object c_get_market_order(self, object market_pair, str order_id):
        return self._tracked_market_orders.get(market_pair, {}).get(order_id)

    def get_market_order(self, market_pair, order_id: str) -> MarketOrder:
        return self.c_get_market_order(market_pair, order_id)

    cdef LimitOrder c_get_shadow_limit_order(self, str order_id):
        cdef:
            object market_pair = self._shadow_order_id_to_market_pair.get(order_id)

        return self._shadow_tracked_limit_orders.get(market_pair, {}).get(order_id)

    def get_shadow_limit_order(self, order_id: str) -> LimitOrder:
        return self.c_get_shadow_limit_order(order_id)

    cdef c_start_tracking_limit_order(self, object market_pair, str order_id, bint is_buy, object price,
                                      object quantity):
        if market_pair not in self._tracked_limit_orders:
            self._tracked_limit_orders[market_pair] = {}
        if market_pair not in self._shadow_tracked_limit_orders:
            self._shadow_tracked_limit_orders[market_pair] = {}

        cdef:
            LimitOrder limit_order = LimitOrder(order_id,
                                                market_pair.trading_pair,
                                                is_buy,
                                                market_pair.base_asset,
                                                market_pair.quote_asset,
                                                price,
                                                quantity,
                                                creation_timestamp=int(self._current_timestamp * 1e6))
        self._tracked_limit_orders[market_pair][order_id] = limit_order
        self._shadow_tracked_limit_orders[market_pair][order_id] = limit_order
        self._order_id_to_market_pair[order_id] = market_pair
        self._shadow_order_id_to_market_pair[order_id] = market_pair

    def start_tracking_limit_order(self, market_pair: MarketTradingPairTuple, order_id: str, is_buy: bool, price: Decimal,
                                   quantity: Decimal):
        return self.c_start_tracking_limit_order(market_pair, order_id, is_buy, price, quantity)

    cdef c_stop_tracking_limit_order(self, object market_pair, str order_id):
        if market_pair in self._tracked_limit_orders and order_id in self._tracked_limit_orders[market_pair]:
            del self._tracked_limit_orders[market_pair][order_id]
            if len(self._tracked_limit_orders[market_pair]) < 1:
                del self._tracked_limit_orders[market_pair]
            self._shadow_gc_requests.append((
                self._current_timestamp + self.SHADOW_MAKER_ORDER_KEEP_ALIVE_DURATION,
                market_pair,
                order_id
            ))

        if order_id in self._order_id_to_market_pair:
            del self._order_id_to_market_pair[order_id]
        if order_id in self._in_flight_cancels:
            del self._in_flight_cancels[order_id]

    def stop_tracking_limit_order(self, market_pair: MarketTradingPairTuple, order_id: str):
        return self.c_stop_tracking_limit_order(market_pair, order_id)

    cdef c_start_tracking_market_order(self, object market_pair, str order_id, bint is_buy, object quantity):
        if market_pair not in self._tracked_market_orders:
            self._tracked_market_orders[market_pair] = {}
        self._tracked_market_orders[market_pair][order_id] = MarketOrder(
            order_id,
            market_pair.trading_pair,
            is_buy,
            market_pair.base_asset,
            market_pair.quote_asset,
            float(quantity),
            self._current_timestamp
        )
        self._order_id_to_market_pair[order_id] = market_pair

    def start_tracking_market_order(self, market_pair: MarketTradingPairTuple, order_id: str, is_buy: bool, quantity: Decimal):
        return self.c_start_tracking_market_order(market_pair, order_id, is_buy, quantity)

    cdef c_stop_tracking_market_order(self, object market_pair, str order_id):
        if market_pair in self._tracked_market_orders and order_id in self._tracked_market_orders[market_pair]:
            del self._tracked_market_orders[market_pair][order_id]
            if len(self._tracked_market_orders[market_pair]) < 1:
                del self._tracked_market_orders[market_pair]
        if order_id in self._order_id_to_market_pair:
            del self._order_id_to_market_pair[order_id]

    def stop_tracking_market_order(self, market_pair: MarketTradingPairTuple, order_id: str):
        return self.c_stop_tracking_market_order(market_pair, order_id)

    cdef c_check_and_cleanup_shadow_records(self):
        cdef:
            double current_timestamp = self._current_timestamp

        while len(self._shadow_gc_requests) > 0 and self._shadow_gc_requests[0][0] < current_timestamp:
            _, market_pair, order_id = self._shadow_gc_requests.popleft()
            if (market_pair in self._shadow_tracked_limit_orders and
                    order_id in self._shadow_tracked_limit_orders[market_pair]):
                del self._shadow_tracked_limit_orders[market_pair][order_id]
                if len(self._shadow_tracked_limit_orders[market_pair]) < 1:
                    del self._shadow_tracked_limit_orders[market_pair]
            if order_id in self._shadow_order_id_to_market_pair:
                del self._shadow_order_id_to_market_pair[order_id]

    def check_and_cleanup_shadow_records(self):
        self.c_check_and_cleanup_shadow_records()

    cdef c_add_create_order_pending(self, str order_id):
        self.in_flight_pending_created.add(order_id)

    def add_create_order_pending(self, order_id: str):
        self.c_add_create_order_pending(order_id)

    cdef c_remove_create_order_pending(self, str order_id):
        self._in_flight_pending_created.discard(order_id)

    def remove_create_order_pending(self, order_id: str):
        self.c_remove_create_order_pending(order_id)



#!/usr/bin/env python

from __future__ import unicode_literals

import sys
import threading
from asyncio import get_event_loop
from contextlib import contextmanager

__all__ = [
    'patch_stdout',
    'StdoutProxy',
]


@contextmanager
def patch_stdout(raw=False, log_field=None):
    proxy = StdoutProxy(raw=raw, log_field=log_field)

    original_stdout = sys.stdout
    original_stderr = sys.stderr

    # Enter.
    sys.stdout = proxy
    sys.stderr = proxy

    try:
        yield
    finally:
        # Exit.
        proxy.flush()

        sys.stdout = original_stdout
        sys.stderr = original_stderr


class StdoutProxy(object):
    """
    Proxy object for stdout which captures everything and prints output inside
    the current application.
    """

    def __init__(self, raw=False, original_stdout=None, log_field=None):
        assert isinstance(raw, bool)
        original_stdout = original_stdout or sys.__stdout__

        self.original_stdout = original_stdout

        self._lock = threading.RLock()
        self._raw = raw
        self._buffer = []

        self.errors = original_stdout.errors
        self.encoding = original_stdout.encoding
        self.log_field = log_field
        self._ev_loop = get_event_loop()

    def _write_and_flush(self, text):
        if not text:
            return

        def write_and_flush():
            self.log_field.log(text)

        def schedule_write_and_flush():
            self._ev_loop.run_in_executor(None, write_and_flush)

        self._ev_loop.call_soon_threadsafe(schedule_write_and_flush)

    def _write(self, data):
        if '\n' in data:
            # When there is a newline in the data, write everything before the newline, including the newline itself.
            before, after = data.rsplit('\n', 1)
            to_write = self._buffer + [before, '\n']
            self._buffer = [after]

            text = ''.join(to_write)
            self._write_and_flush(text)
        else:
            # Otherwise, cache in buffer.
            self._buffer.append(data)

    def _flush(self):
        text = ''.join(self._buffer)
        self._buffer = []
        self._write_and_flush(text)

    def write(self, data):
        with self._lock:
            self._write(data)

    def flush(self):
        """
        Flush buffered output.
        """
        with self._lock:
            self._flush()

    @staticmethod
    def isatty() -> bool:
        return False

    def fileno(self) -> int:
        return self.original_stdout.fileno()



#!/usr/bin/env python

from sqlalchemy import JSON, BigInteger, Column, Index, Integer, Text

from . import HummingbotBase


class MarketState(HummingbotBase):
    __tablename__ = "MarketState"
    __table_args = (Index("ms_config_market_index",
                          "config_file_path", "market", unique=True),)

    id = Column(Integer, primary_key=True, nullable=False)
    config_file_path = Column(Text, nullable=False)
    market = Column(Text, nullable=False)
    timestamp = Column(BigInteger, nullable=False)
    saved_state = Column(JSON, nullable=False)

    def __repr__(self) -> str:
        return f"MarketState(id='{self.id}', config_file_path='{self.config_file_path}', market='{self.market}', " \
            f"timestamp={self.timestamp}, saved_state={self.saved_state})"



"""Transaction."""

import re
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Optional, Union

from google.protobuf.any_pb2 import Any as ProtoAny
from v4_proto.cosmos.base.v1beta1.coin_pb2 import Coin
from v4_proto.cosmos.crypto.secp256k1.keys_pb2 import PubKey as ProtoPubKey
from v4_proto.cosmos.tx.signing.v1beta1.signing_pb2 import SignMode
from v4_proto.cosmos.tx.v1beta1.tx_pb2 import AuthInfo, Fee, ModeInfo, SignDoc, SignerInfo, Tx, TxBody

from hummingbot.connector.derivative.dydx_v4_perpetual.data_sources.keypairs import PublicKey


def parse_coins(value: str) -> List[Coin]:
    """Parse the coins.

    :param value: coins
    :raises RuntimeError: If unable to parse the value
    :return: coins
    """
    coins = []

    parts = re.split(r",\s*", value)
    for part in parts:
        part = part.strip()
        if part == "":
            continue

        match = re.match(r"(\d+)(\w+)", part)
        if match is None:
            raise RuntimeError(f"Unable to parse value {part}")

        # extract out the groups
        amount, denom = match.groups()
        coins.append(Coin(amount=amount, denom=denom))

    return coins


class TxState(Enum):
    """Transaction state.

    :param Enum: Draft, Sealed, Final
    """

    Draft = 0
    Sealed = 1
    Final = 2


def _is_iterable(value) -> bool:
    try:
        iter(value)
        return True
    except TypeError:
        return False


def _wrap_in_proto_any(values: List[Any]) -> List[ProtoAny]:
    any_values = []
    for value in values:
        proto_any = ProtoAny()
        proto_any.Pack(value, type_url_prefix="/")  # type: ignore
        any_values.append(proto_any)
    return any_values


def _create_proto_public_key(public_key: PublicKey) -> ProtoAny:
    proto_public_key = ProtoAny()
    proto_public_key.Pack(
        ProtoPubKey(
            key=public_key.public_key_bytes,
        ),
        type_url_prefix="/",
    )
    return proto_public_key


class SigningMode(Enum):
    """Signing mode.

    :param Enum: Direct
    """

    Direct = 1


@dataclass
class SigningCfg:
    """Transaction signing configuration."""

    mode: SigningMode
    sequence_num: int
    public_key: PublicKey

    @staticmethod
    def direct(public_key: PublicKey, sequence_num: int) -> "SigningCfg":
        """Transaction signing configuration using direct mode.

        :param public_key: public key
        :param sequence_num: sequence number
        :return: Transaction signing configuration
        """
        return SigningCfg(
            mode=SigningMode.Direct,
            sequence_num=sequence_num,
            public_key=public_key,
        )


class Transaction:
    """Transaction."""

    def __init__(self):
        """Init the Transactions with transaction message, state, fee and body."""
        self._msgs: List[Any] = []
        self._state: TxState = TxState.Draft
        self._tx_body: Optional[TxBody] = None
        self._tx = None
        self._fee = None

    @property  # noqa
    def state(self) -> TxState:
        """Get the transaction state.

        :return: current state of the transaction
        """
        return self._state

    @property  # noqa
    def msgs(self):
        """Get the transaction messages.

        :return: transaction messages
        """
        return self._msgs

    @property
    def fee(self) -> Optional[str]:
        """Get the transaction fee.

        :return: transaction fee
        """
        return self._fee

    @property
    def tx(self):
        """Initialize.

        :raises RuntimeError: If the transaction has not been completed.
        :return: transaction
        """
        if self._state != TxState.Final:
            raise RuntimeError("The transaction has not been completed")
        return self._tx

    def add_message(self, msg: Any) -> "Transaction":
        """Initialize.

        :param msg: transaction message (memo)
        :raises RuntimeError: If the transaction is not in the draft state.
        :return: transaction with message added
        """
        if self._state != TxState.Draft:
            raise RuntimeError(
                "The transaction is not in the draft state. No further messages may be appended"
            )
        self._msgs.append(msg)
        return self

    def seal(
            self,
            signing_cfgs: Union[SigningCfg, List[SigningCfg]],
            fee: str,
            gas_limit: int,
            memo: Optional[str] = None,
    ) -> "Transaction":
        """Seal the transaction.

        :param signing_cfgs: signing configs
        :param fee: transaction fee
        :param gas_limit: transaction gas limit
        :param memo: transaction memo, defaults to None
        :return: sealed transaction.
        """
        self._state = TxState.Sealed

        input_signing_cfgs: List[SigningCfg] = (
            signing_cfgs if _is_iterable(signing_cfgs) else [signing_cfgs]  # type: ignore
        )

        signer_infos = []
        for signing_cfg in input_signing_cfgs:
            assert signing_cfg.mode == SigningMode.Direct

            signer_infos.append(
                SignerInfo(
                    public_key=_create_proto_public_key(signing_cfg.public_key),
                    mode_info=ModeInfo(
                        single=ModeInfo.Single(mode=SignMode.SIGN_MODE_DIRECT)
                    ),
                    sequence=signing_cfg.sequence_num,
                )
            )

        auth_info = AuthInfo(
            signer_infos=signer_infos,
            fee=Fee(amount=parse_coins(fee), gas_limit=gas_limit),
        )

        self._fee = fee

        self._tx_body = TxBody()
        self._tx_body.memo = memo or ""
        self._tx_body.messages.extend(
            _wrap_in_proto_any(self._msgs)
        )  # pylint: disable=E1101

        self._tx = Tx(body=self._tx_body, auth_info=auth_info)
        return self

    def sign(
            self,
            signer,
            chain_id: str,
            account_number: int,
            deterministic: bool = False,
    ) -> "Transaction":
        """Sign the transaction.

        :param signer: Signer
        :param chain_id: chain id
        :param account_number: account number
        :param deterministic: deterministic, defaults to False
        :raises RuntimeError: If transaction is not sealed
        :return: signed transaction
        """
        if self.state != TxState.Sealed:
            raise RuntimeError(
                "Transaction is not sealed. It must be sealed before signing is possible."
            )

        sd = SignDoc()
        sd.body_bytes = self._tx.body.SerializeToString()
        sd.auth_info_bytes = self._tx.auth_info.SerializeToString()
        sd.chain_id = chain_id
        sd.account_number = account_number

        data_for_signing = sd.SerializeToString()

        # Generating deterministic signature:
        signature = signer.sign(
            data_for_signing,
            deterministic=deterministic,
            canonicalise=True,
        )
        self._tx.signatures.extend([signature])
        return self

    def complete(self) -> "Transaction":
        """Update transaction state to Final.

        :return: transaction with  updated state
        """
        self._state = TxState.Final
        return self



from hummingbot.connector.exchange.mexc.protobuf import PublicBookTickerV3Api_pb2 as _PublicBookTickerV3Api_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PublicBookTickerBatchV3Api(_message.Message):
    __slots__ = ("items",)
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[_PublicBookTickerV3Api_pb2.PublicBookTickerV3Api]
    def __init__(self, items: _Optional[_Iterable[_Union[_PublicBookTickerV3Api_pb2.PublicBookTickerV3Api, _Mapping]]] = ...) -> None: ...



import asyncio
import hashlib
from decimal import Decimal
from typing import Any, AsyncIterable, Dict, List, Literal, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.hyperliquid import (
    hyperliquid_constants as CONSTANTS,
    hyperliquid_web_utils as web_utils,
)
from hummingbot.connector.exchange.hyperliquid.hyperliquid_api_order_book_data_source import (
    HyperliquidAPIOrderBookDataSource,
)
from hummingbot.connector.exchange.hyperliquid.hyperliquid_api_user_stream_data_source import (
    HyperliquidAPIUserStreamDataSource,
)
from hummingbot.connector.exchange.hyperliquid.hyperliquid_auth import HyperliquidAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import TradeFillOrderDetails, combine_to_hb_trading_pair, get_new_client_order_id
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import DeductedFromReturnsTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.event.events import MarketEvent, OrderFilledEvent
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class HyperliquidExchange(ExchangePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0

    web_utils = web_utils

    SHORT_POLL_INTERVAL = 5.0
    LONG_POLL_INTERVAL = 120.0

    def __init__(
            self,
            balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
            rate_limits_share_pct: Decimal = Decimal("100"),
            hyperliquid_secret_key: str = None,
            hyperliquid_address: str = None,
            use_vault: bool = False,
            hyperliquid_mode: Literal["arb_wallet", "api_wallet"] = "arb_wallet",
            trading_pairs: Optional[List[str]] = None,
            trading_required: bool = True,
            domain: str = CONSTANTS.DOMAIN,
    ):
        self.hyperliquid_address = hyperliquid_address
        self.hyperliquid_secret_key = hyperliquid_secret_key
        self._use_vault = use_vault
        self._connection_mode = hyperliquid_mode
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._domain = domain
        self._last_trade_history_timestamp = None
        self._last_trades_poll_timestamp = 1.0
        self.coin_to_asset: Dict[str, int] = {}
        self.name_to_coin: Dict[str, str] = {}
        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def name(self) -> str:
        # Note: domain here refers to the entire exchange name. i.e. hyperliquid_ or hyperliquid_testnet
        return self._domain

    @property
    def authenticator(self) -> Optional[HyperliquidAuth]:
        if self._trading_required:
            return HyperliquidAuth(
                self.hyperliquid_address,
                self.hyperliquid_secret_key,
                self._use_vault
            )
        return None

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.BROKER_ID

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.PING_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    async def _make_network_check_request(self):
        await self._api_post(path_url=self.check_network_request_path, data={"type": CONSTANTS.META_INFO})

    def supported_order_types(self) -> List[OrderType]:
        """
        :return a list of OrderType supported by this connector
        """
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    async def get_all_pairs_prices(self) -> List[Dict[str, str]]:
        res = []
        exchange_info = await self._api_post(
            path_url=CONSTANTS.TICKER_PRICE_CHANGE_URL,
            data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        spot_infos: list = exchange_info[1]
        for spot_data in spot_infos:
            res.append({
                "symbol": spot_data.get("coin"),
                "price": spot_data.get("markPx"),
            })

        return res

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            auth=self._auth)

    async def _make_trading_rules_request(self) -> Any:
        exchange_info = await self._api_post(path_url=self.trading_rules_request_path,
                                             data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        return exchange_info

    async def _make_trading_pairs_request(self) -> Any:
        exchange_info = await self._api_post(path_url=self.trading_pairs_request_path,
                                             data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        return exchange_info

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return CONSTANTS.UNKNOWN_ORDER_MESSAGE in str(cancelation_exception)

    def quantize_order_price(self, trading_pair: str, price: Decimal) -> Decimal:
        """
        Applies trading rule to quantize order price.
        """
        d_price = Decimal(round(float(f"{price:.5g}"), 6))
        return d_price

    async def _update_trading_rules(self):
        exchange_info = await self._api_post(path_url=self.trading_rules_request_path,
                                             data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        trading_rules_list = await self._format_trading_rules(exchange_info)
        self._trading_rules.clear()
        for trading_rule in trading_rules_list:
            self._trading_rules[trading_rule.trading_pair] = trading_rule
        self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)

    async def _initialize_trading_pair_symbol_map(self):
        try:
            exchange_info = await self._api_post(path_url=self.trading_pairs_request_path,
                                                 data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})

            self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)
        except Exception:
            self.logger().exception("There was an error requesting exchange info.")

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return HyperliquidAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return HyperliquidAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    async def _status_polling_loop_fetch_updates(self):
        await safe_gather(
            self._update_trade_history(),
            self._update_order_status(),
            self._update_balances(),
        )

    async def _update_order_status(self):
        await self._update_orders()

    async def _update_lost_orders_status(self):
        await self._update_lost_orders()

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = order_type is OrderType.LIMIT_MAKER
        return DeductedFromReturnsTradeFee(percent=self.estimate_fee_pct(is_maker))

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        api_params = {
            "type": "cancel",
            "cancels": {
                "asset": self.coin_to_asset[symbol],
                "cloid": order_id
            },
        }
        cancel_result = await self._api_post(
            path_url=CONSTANTS.CANCEL_ORDER_URL,
            data=api_params,
            is_auth_required=True)

        if cancel_result.get("status") == "err" or "error" in cancel_result["response"]["data"]["statuses"][0]:
            self.logger().debug(f"The order {order_id} does not exist on Hyperliquid s. "
                                f"No cancelation needed.")
            await self._order_tracker.process_order_not_found(order_id)
            raise IOError(f'{cancel_result["response"]}')
        if cancel_result["status"] == "ok" and "success" in cancel_result["response"]["data"]["statuses"][0]:
            return True
        return False

    # === Orders placing ===

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        """
        Creates a promise to create a buy order using the parameters

        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price

        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=True,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.md5()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"
        if order_type is OrderType.MARKET:
            reference_price = self.get_mid_price(trading_pair) if price.is_nan() else price
            price = self.quantize_order_price(trading_pair, reference_price * Decimal(1 + CONSTANTS.MARKET_ORDER_SLIPPAGE))

        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        """
        Creates a promise to create a sell order using the parameters.
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=False,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.md5()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"
        if order_type is OrderType.MARKET:
            reference_price = self.get_mid_price(trading_pair) if price.is_nan() else price
            price = self.quantize_order_price(trading_pair, reference_price * Decimal(1 - CONSTANTS.MARKET_ORDER_SLIPPAGE))

        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    async def _place_order(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Decimal,
            **kwargs,
    ) -> Tuple[str, float]:

        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        param_order_type = {"limit": {"tif": "Gtc"}}
        if order_type is OrderType.LIMIT_MAKER:
            param_order_type = {"limit": {"tif": "Alo"}}
        if order_type is OrderType.MARKET:
            param_order_type = {"limit": {"tif": "Ioc"}}

        api_params = {
            "type": "order",
            "grouping": "na",
            "orders": {
                "asset": self.coin_to_asset[symbol],
                "isBuy": True if trade_type is TradeType.BUY else False,
                "limitPx": float(price),
                "sz": float(amount),
                "reduceOnly": False,
                "orderType": param_order_type,
                "cloid": order_id,
            }
        }
        order_result = await self._api_post(
            path_url = CONSTANTS.CREATE_ORDER_URL,
            data = api_params,
            is_auth_required = True)
        if order_result.get("status") == "err":
            raise IOError(f"Error submitting order {order_id}: {order_result['response']}")
        else:
            o_order_result = order_result['response']["data"]["statuses"][0]
        if "error" in o_order_result:
            raise IOError(f"Error submitting order {order_id}: {o_order_result['error']}")
        o_data = o_order_result.get("resting") or o_order_result.get("filled")
        o_id = str(o_data["oid"])
        return (o_id, self.current_timestamp)

    async def _update_trade_history(self):
        orders = list(self._order_tracker.all_fillable_orders.values())
        all_fillable_orders = self._order_tracker.all_fillable_orders_by_exchange_order_id
        all_fills_response = []
        if len(orders) > 0:
            try:
                all_fills_response = await self._api_post(
                    path_url = CONSTANTS.ACCOUNT_TRADE_LIST_URL,
                    data = {
                        "type": CONSTANTS.TRADES_TYPE,
                        "user": self.hyperliquid_address,
                    })
            except asyncio.CancelledError:
                raise
            except Exception as request_error:
                self.logger().warning(
                    f"Failed to fetch trade updates. Error: {request_error}",
                    exc_info = request_error,
                )
            for trade_fill in all_fills_response:
                self._process_trade_rs_event_message(order_fill=trade_fill, all_fillable_order=all_fillable_orders)

    def _process_trade_rs_event_message(self, order_fill: Dict[str, Any], all_fillable_order):
        exchange_order_id = str(order_fill.get("oid"))
        fillable_order = all_fillable_order.get(exchange_order_id)
        if fillable_order is not None:
            fee_asset = order_fill["feeToken"]

            fee = TradeFeeBase.new_spot_fee(
                fee_schema=self.trade_fee_schema(),
                trade_type=fillable_order.trade_type,
                percent_token=fee_asset,
                flat_fees=[TokenAmount(amount=Decimal(order_fill["fee"]), token=fee_asset)]
            )

            trade_update = TradeUpdate(
                trade_id=str(order_fill["tid"]),
                client_order_id=fillable_order.client_order_id,
                exchange_order_id=str(order_fill["oid"]),
                trading_pair=fillable_order.trading_pair,
                fee=fee,
                fill_base_amount=Decimal(order_fill["sz"]),
                fill_quote_amount=Decimal(order_fill["px"]) * Decimal(order_fill["sz"]),
                fill_price=Decimal(order_fill["px"]),
                fill_timestamp=order_fill["time"] * 1e-3,
            )

            self._order_tracker.process_trade_update(trade_update)

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, any]]:
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unknown error. Retrying after 1 seconds.",
                    exc_info=True,
                    app_warning_msg="Could not fetch user events from Hyperliquid. Check API key and network connection.",
                )
                await self._sleep(1.0)

    async def _user_stream_event_listener(self):
        """
        Listens to messages from _user_stream_tracker.user_stream queue.
        Traders, Orders, and Balance updates from the WS.
        """
        user_channels = [
            CONSTANTS.USER_ORDERS_ENDPOINT_NAME,
            CONSTANTS.USEREVENT_ENDPOINT_NAME,
        ]
        async for event_message in self._iter_user_event_queue():
            try:
                if isinstance(event_message, dict):
                    channel: str = event_message.get("channel", None)
                    results = event_message.get("data", None)
                elif event_message is asyncio.CancelledError:
                    raise asyncio.CancelledError
                else:
                    raise Exception(event_message)
                if channel not in user_channels:
                    self.logger().error(
                        f"Unexpected message in user stream: {event_message}.", exc_info=True)
                    continue
                if channel == CONSTANTS.USER_ORDERS_ENDPOINT_NAME:
                    for order_msg in results:
                        self._process_order_message(order_msg)
                elif channel == CONSTANTS.USEREVENT_ENDPOINT_NAME:
                    if "fills" in results:
                        for trade_msg in results["fills"]:
                            client_order_id = str(trade_msg.get("cloid", ""))
                            await self._process_trade_message(trade_msg, client_order_id)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    async def _process_trade_message(self, trade: Dict[str, Any], client_order_id: Optional[str] = None):
        """
        Updates in-flight order and trigger order filled event for a trade message received. Triggers order completedim
        event if the total executed amount equals to the specified order amount.
        Example Trade:
        """
        tracked_order = self._order_tracker.all_fillable_orders.get(client_order_id)

        if tracked_order is not None:
            trading_pair_base_coin = tracked_order.trading_pair
            exchange_symbol = await self.trading_pair_associated_to_exchange_symbol(symbol=trade["coin"])
            if exchange_symbol == trading_pair_base_coin:
                fee_asset = trade["feeToken"]
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=tracked_order.trade_type,
                    percent_token=fee_asset,
                    flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=fee_asset)]
                )
                trade_update: TradeUpdate = TradeUpdate(
                    trade_id=str(trade["tid"]),
                    client_order_id=tracked_order.client_order_id,
                    exchange_order_id=str(trade["oid"]),
                    trading_pair=tracked_order.trading_pair,
                    fill_timestamp=trade["time"] * 1e-3,
                    fill_price=Decimal(trade["px"]),
                    fill_base_amount=Decimal(trade["sz"]),
                    fill_quote_amount=Decimal(trade["px"]) * Decimal(trade["sz"]),
                    fee=fee,
                )
                self._order_tracker.process_trade_update(trade_update)

    def _process_order_message(self, order_msg: Dict[str, Any]):
        """
        Updates in-flight order and triggers cancelation or failure event if needed.

        :param order_msg: The order response from either REST or web socket API (they are of the same format)

        Example Order:
        """
        client_order_id = str(order_msg["order"].get("cloid", ""))
        tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
        if not tracked_order:
            self.logger().debug(f"Ignoring order message with id {client_order_id}: not in in_flight_orders.")
            return
        current_state = order_msg["status"]
        order_update: OrderUpdate = OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            update_timestamp=order_msg["statusTimestamp"] * 1e-3,
            new_state=CONSTANTS.ORDER_STATE[current_state],
            client_order_id=order_msg["order"]["cloid"],
            exchange_order_id=str(order_msg["order"]["oid"]),
        )
        self._order_tracker.process_order_update(order_update=order_update)

    async def _format_trading_rules(self, exchange_info_dict: List) -> List[TradingRule]:
        """
        Queries the necessary API endpoint and initialize the TradingRule object for each trading pair being traded.

        Parameters
        ----------
        exchange_info_dict:
            Trading rules dictionary response from the exchange
        """
        self.coin_to_asset = {}
        self.name_to_coin = {}

        self.coin_to_asset = {asset_info["name"]: asset for (asset, asset_info) in enumerate(exchange_info_dict[0]["universe"])}
        self.name_to_coin = {asset_info["name"]: asset_info["name"] for asset_info in exchange_info_dict[0]["universe"]}

        coin_infos: list = exchange_info_dict[0]["universe"]
        price_infos: list = exchange_info_dict[1]

        for spot_info in filter(web_utils.is_exchange_information_valid, exchange_info_dict[0]["universe"]):
            self.coin_to_asset[spot_info["name"]] = spot_info["index"] + 10000
            self.name_to_coin[spot_info["name"]] = spot_info["name"]

        return_val: list = []
        for coin_info, price_info in zip(coin_infos, price_infos):
            base, quote = coin_info["tokens"]
            try:
                ex_name = f'{exchange_info_dict[0]["tokens"][base]["name"].replace(" ", "").upper()}/{exchange_info_dict[0]["tokens"][quote]["name"].replace(" ", "").upper()}'
                if ex_name not in self.name_to_coin:
                    self.name_to_coin[ex_name] = coin_info["name"]

                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=coin_info["name"])
                step_size = Decimal(str(10 ** -exchange_info_dict[0]["tokens"][base].get("szDecimals")))
                price_size = Decimal(str(10 ** -len(price_info.get("markPx").split('.')[1])))
                return_val.append(
                    TradingRule(
                        trading_pair,
                        min_order_size=step_size,  # asset_price,
                        min_base_amount_increment=step_size,
                        min_price_increment=price_size
                    )
                )
            except Exception:
                self.logger().error(f"Error parsing the trading pair rule {exchange_info_dict}. Skipping.",
                                    exc_info=True)
        return return_val

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List):
        mapping = bidict()
        self.coin_to_asset = {}
        self.name_to_coin = {}

        self.coin_to_asset = {asset_info["name"]: asset for (asset, asset_info) in enumerate(exchange_info[0]["universe"])}

        self.name_to_coin = {asset_info["name"]: asset_info["name"] for asset_info in exchange_info[0]["universe"]}

        for spot_info in filter(web_utils.is_exchange_information_valid, exchange_info[0]["universe"]):
            self.coin_to_asset[spot_info["name"]] = spot_info["index"] + 10000
            self.name_to_coin[spot_info["name"]] = spot_info["name"]
            base, quote = spot_info["tokens"]
            name = f'{exchange_info[0]["tokens"][base]["name"].replace(" ", "").upper()}/{exchange_info[0]["tokens"][quote]["name"].replace(" ", "").upper()}'
            ex_name = spot_info["name"]

            if name not in self.name_to_coin:
                self.name_to_coin[name] = spot_info["name"]

            new_base, new_quote = name.split("/")
            trading_pair = combine_to_hb_trading_pair(new_base, new_quote)
            if trading_pair in mapping.inverse:
                self._resolve_trading_pair_symbols_duplicate(mapping, name, new_base, new_quote)
            else:
                mapping[ex_name] = trading_pair
        self._set_trading_pair_symbol_map(mapping)

    def _resolve_trading_pair_symbols_duplicate(self, mapping: bidict, new_exchange_symbol: str, base: str, quote: str):
        """Resolves name conflicts provoked by futures contracts.

        If the expected BASEQUOTE combination matches one of the exchange symbols, it is the one taken, otherwise,
        the trading pair is removed from the map and an error is logged.
        """
        expected_exchange_symbol = f"{base}{quote}"
        trading_pair = combine_to_hb_trading_pair(base, quote)
        current_exchange_symbol = mapping.inverse[trading_pair]
        if current_exchange_symbol == expected_exchange_symbol:
            pass
        elif new_exchange_symbol == expected_exchange_symbol:
            mapping.pop(current_exchange_symbol)
            mapping[new_exchange_symbol] = trading_pair
        else:
            self.logger().error(
                f"Could not resolve the exchange symbols {new_exchange_symbol} and {current_exchange_symbol}")
            mapping.pop(current_exchange_symbol)

    async def _update_balances(self):
        """
        Calls the REST API to update total and available balances.
        """
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        account_info = await self._api_post(path_url=CONSTANTS.ACCOUNT_INFO_URL,
                                            data={"type": CONSTANTS.USER_STATE_TYPE,
                                                  "user": self.hyperliquid_address},
                                            )
        balances = account_info["balances"]
        for balance_entry in balances:
            asset_name = balance_entry["coin"]
            free_balance = Decimal(balance_entry["total"]) - Decimal(balance_entry["hold"])
            total_balance = Decimal(balance_entry["total"])
            self._account_available_balances[asset_name] = free_balance
            self._account_balances[asset_name] = total_balance
            remote_asset_names.add(asset_name)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        client_order_id = tracked_order.client_order_id
        order_update = await self._api_post(
            path_url=CONSTANTS.ORDER_URL,
            data={
                "type": CONSTANTS.ORDER_STATUS_TYPE,
                "user": self.hyperliquid_address,
                "oid": int(tracked_order.exchange_order_id) if tracked_order.exchange_order_id else client_order_id
            })
        current_state = order_update["order"]["status"]
        _order_update: OrderUpdate = OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            update_timestamp=order_update["order"]["order"]["timestamp"] * 1e-3,
            new_state=CONSTANTS.ORDER_STATE[current_state],
            client_order_id=order_update["order"]["order"]["cloid"] or client_order_id,
            exchange_order_id=str(tracked_order.exchange_order_id),
        )
        return _order_update

    async def _update_order_fills_from_trades(self):
        """
        This is intended to be a backup measure to get filled events with trade ID for orders,
        in case hyperliquid's user stream events are not working.
        NOTE: It is not required to copy this functionality in other connectors.
        This is separated from _update_order_status which only updates the order status without producing filled
        events, since hyperliquid's get order endpoint does not return trade IDs.
        The minimum poll interval for order status is 10 seconds.
        """
        small_interval_last_tick = self._last_poll_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL
        small_interval_current_tick = self.current_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL
        long_interval_last_tick = self._last_poll_timestamp / self.LONG_POLL_INTERVAL
        long_interval_current_tick = self.current_timestamp / self.LONG_POLL_INTERVAL

        if (long_interval_current_tick > long_interval_last_tick
                or (self.in_flight_orders and small_interval_current_tick > small_interval_last_tick)):
            query_time = int(self._last_trades_poll_timestamp * 1e3)
            self._last_trades_poll_timestamp = self._time_synchronizer.time()
            order_by_exchange_id_map = {}
            for order in self._order_tracker.all_fillable_orders.values():
                order_by_exchange_id_map[order.exchange_order_id] = order

            tasks = []
            trading_pairs = self.trading_pairs
            for trading_pair in trading_pairs:
                params = {
                    'type': CONSTANTS.TRADES_TYPE,
                    'user': self.hyperliquid_address,
                }
                if self._last_poll_timestamp > 0:
                    params['type'] = 'userFillsByTime'
                    params["startTime"] = query_time
                tasks.append(self._api_get(
                    path_url=CONSTANTS.MY_TRADES_PATH_URL,
                    params=params,
                    is_auth_required=True))

            self.logger().debug(f"Polling for order fills of {len(tasks)} trading pairs.")
            results = await safe_gather(*tasks, return_exceptions=True)

            for trades, trading_pair in zip(results, trading_pairs):

                if isinstance(trades, Exception):
                    self.logger().network(
                        f"Error fetching trades update for the order {trading_pair}: {trades}.",
                        app_warning_msg=f"Failed to fetch trade update for {trading_pair}."
                    )
                    continue
                for trade in trades:
                    exchange_order_id = str(trade["oid"])
                    if exchange_order_id in order_by_exchange_id_map:
                        # This is a fill for a tracked order
                        tracked_order = order_by_exchange_id_map[exchange_order_id]
                        fee = TradeFeeBase.new_spot_fee(
                            fee_schema=self.trade_fee_schema(),
                            trade_type=tracked_order.trade_type,
                            percent_token=trade["feeToken"],
                            flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=trade["feeToken"])]
                        )
                        trade_update = TradeUpdate(
                            trade_id=str(trade["tid"]),
                            client_order_id=tracked_order.client_order_id,
                            exchange_order_id=exchange_order_id,
                            trading_pair=trading_pair,
                            fee=fee,
                            fill_base_amount=Decimal(trade["sz"]),
                            fill_quote_amount=Decimal(trade["sz"]) * Decimal(trade["px"]),
                            fill_price=Decimal(trade["px"]),
                            fill_timestamp=trade["time"] * 1e-3,
                        )
                        self._order_tracker.process_trade_update(trade_update)
                    elif self.is_confirmed_new_order_filled_event(str(trade["tid"]), exchange_order_id, trading_pair):
                        # This is a fill of an order registered in the DB but not tracked any more
                        self._current_trade_fills.add(TradeFillOrderDetails(
                            market=self.display_name,
                            exchange_trade_id=str(trade["tid"]),
                            symbol=trading_pair))
                        self.trigger_event(
                            MarketEvent.OrderFilled,
                            OrderFilledEvent(
                                timestamp=float(trade["time"]) * 1e-3,
                                order_id=self._exchange_order_ids.get(str(trade["oid"]), None),
                                trading_pair=trading_pair,
                                trade_type=TradeType.BUY if trade["side"] == 'B' else TradeType.SELL,
                                order_type=OrderType.LIMIT_MAKER if "Open" in trade["dir"] else OrderType.LIMIT,
                                price=Decimal(trade["px"]),
                                amount=Decimal(trade["sz"]),
                                trade_fee=DeductedFromReturnsTradeFee(
                                    flat_fees=[
                                        TokenAmount(
                                            trade["feeToken"],
                                            Decimal(trade["fee"])
                                        )
                                    ]
                                ),
                                exchange_trade_id=str(trade["tid"])
                            ))
                        self.logger().info(f"Recreating missing trade in TradeFill: {trade}")

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = int(order.exchange_order_id)
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            all_fills_response = await self._api_get(
                path_url=CONSTANTS.MY_TRADES_PATH_URL,
                params={
                    "type": "userFills",
                    'user': self.hyperliquid_address,
                },
                is_auth_required=True,
                limit_id=CONSTANTS.MY_TRADES_PATH_URL)

            for trade in all_fills_response:
                exchange_order_id = str(trade["orderId"])
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=order.trade_type,
                    percent_token=trade["feeToken"],
                    flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=trade["feeToken"])]
                )
                trade_update = TradeUpdate(
                    trade_id=str(trade["tid"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=exchange_order_id,
                    trading_pair=trading_pair,
                    fee=fee,
                    fill_base_amount=Decimal(trade["sz"]),
                    fill_quote_amount=Decimal(trade["sz"]) * Decimal(trade["px"]),
                    fill_price=Decimal(trade["px"]),
                    fill_timestamp=trade["time"] * 1e-3,
                )
                trade_updates.append(trade_update)

        return trade_updates

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        response = await self._api_post(path_url=CONSTANTS.TICKER_PRICE_CHANGE_URL,
                                        data={"type": CONSTANTS.ASSET_CONTEXT_TYPE})
        price = 0
        for token in response[1]:
            if token['coin'] == exchange_symbol:
                price = token['markPx']
                break
        return price



from typing import Callable, Optional
from urllib.parse import urljoin

from hummingbot.connector.derivative.bitget_perpetual import bitget_perpetual_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import TimeSynchronizerRESTPreProcessor
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def public_ws_url(domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public websocket endpoint
    """
    return _create_ws_url(CONSTANTS.WSS_PUBLIC_ENDPOINT, domain)


def private_ws_url(domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided private websocket endpoint
    """
    return _create_ws_url(CONSTANTS.WSS_PRIVATE_ENDPOINT, domain)


def public_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public REST endpoint

    :param path_url: a public REST endpoint
    :param domain: the Bitget domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    return _create_rest_url(path_url, domain)


def private_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided private REST endpoint

    :param path_url: a private REST endpoint
    :param domain: the Bitget domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    return _create_rest_url(path_url, domain)


def _create_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided REST endpoint

    :param path_url: a REST endpoint
    :param domain: the Bitget domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    return urljoin(f"https://{CONSTANTS.REST_SUBDOMAIN}.{domain}", path_url)


def _create_ws_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided websocket endpoint

    :param path_url: a websocket endpoint
    :param domain: the Bitget domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    return urljoin(f"wss://{CONSTANTS.WSS_SUBDOMAIN}.{domain}", path_url)


def build_api_factory(
        throttler: Optional[AsyncThrottler] = None,
        time_synchronizer: Optional[TimeSynchronizer] = None,
        time_provider: Optional[Callable] = None,
        auth: Optional[AuthBase] = None,
) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    time_synchronizer = time_synchronizer or TimeSynchronizer()
    time_provider = time_provider or (lambda: get_current_server_time(throttler=throttler))
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        rest_pre_processors=[
            TimeSynchronizerRESTPreProcessor(
                synchronizer=time_synchronizer,
                time_provider=time_provider
            ),
        ],
    )

    return api_factory


def build_api_factory_without_time_synchronizer_pre_processor(
    throttler: AsyncThrottler
) -> WebAssistantsFactory:
    """
    Build an API factory without the time synchronizer pre-processor.

    :param throttler: The throttler to use for the API factory.
    :return: The API factory.
    """
    api_factory = WebAssistantsFactory(throttler=throttler)

    return api_factory


def create_throttler() -> AsyncThrottler:
    """
    Create a throttler with the default rate limits.

    :return: The throttler.
    """
    throttler = AsyncThrottler(CONSTANTS.RATE_LIMITS)

    return throttler


async def get_current_server_time(
    throttler: Optional[AsyncThrottler] = None,
    domain: str = CONSTANTS.DEFAULT_DOMAIN
) -> float:
    """
    Get the current server time in seconds.

    :param throttler: The throttler to use for the request.
    :param domain: The domain to use for the request.
    :return: The current server time in seconds.
    """
    throttler = throttler or create_throttler()
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    rest_assistant = await api_factory.get_rest_assistant()

    url = public_rest_url(path_url=CONSTANTS.PUBLIC_TIME_ENDPOINT, domain=domain)
    response = await rest_assistant.execute_request(
        url=url,
        throttler_limit_id=CONSTANTS.PUBLIC_TIME_ENDPOINT,
        method=RESTMethod.GET,
        return_err=True,
    )
    server_time = float(response["requestTime"])

    return server_time



from asyncio import Lock
from datetime import datetime, timedelta
from typing import Optional, Tuple

import certifi
import grpc
from google.protobuf import json_format, message as _message
from v4_proto.cosmos.auth.v1beta1.auth_pb2 import BaseAccount
from v4_proto.cosmos.auth.v1beta1.query_pb2 import QueryAccountRequest
from v4_proto.cosmos.auth.v1beta1.query_pb2_grpc import QueryStub as AuthGrpcClient
from v4_proto.cosmos.bank.v1beta1 import query_pb2_grpc as bank_query_grpc
from v4_proto.cosmos.base.tendermint.v1beta1 import (
    query_pb2 as tendermint_query,
    query_pb2_grpc as tendermint_query_grpc,
)
from v4_proto.cosmos.tx.v1beta1.service_pb2 import BroadcastMode, BroadcastTxRequest
from v4_proto.cosmos.tx.v1beta1.service_pb2_grpc import ServiceStub as TxGrpcClient
from v4_proto.dydxprotocol.clob.order_pb2 import Order, OrderId
from v4_proto.dydxprotocol.clob.tx_pb2 import MsgCancelOrder, MsgPlaceOrder
from v4_proto.dydxprotocol.subaccounts.subaccount_pb2 import SubaccountId

from hummingbot.connector.derivative.dydx_v4_perpetual import dydx_v4_perpetual_constants as CONSTANTS
from hummingbot.connector.derivative.dydx_v4_perpetual.data_sources.keypairs import PrivateKey
from hummingbot.connector.derivative.dydx_v4_perpetual.data_sources.tx import SigningCfg, Transaction


class DydxPerpetualV4Client:

    def __init__(
            self,
            secret_phrase: str,
            dydx_v4_chain_address: str,
            connector,
            subaccount_num=0,
    ):
        self._private_key = PrivateKey.from_mnemonic(secret_phrase)
        self._dydx_v4_chain_address = dydx_v4_chain_address
        self._connector = connector
        self._subaccount_num = subaccount_num
        self.transaction_lock = Lock()
        self.number = 0
        self.sequence = 0
        self._is_trading_account_initialized = False

        with open(certifi.where(), "rb") as f:
            trusted_certs = f.read()
        credentials = grpc.ssl_channel_credentials(
            root_certificates=trusted_certs
        )

        host_and_port = CONSTANTS.DYDX_V4_AERIAL_CONFIG_URL
        grpc_client = (
            grpc.aio.secure_channel(host_and_port, credentials)
            if credentials is not None else grpc.aio.insecure_channel(host_and_port)
        )
        query_grpc_client = (
            grpc.aio.secure_channel(CONSTANTS.DYDX_V4_QUERY_AERIAL_CONFIG_URL, credentials)
            if credentials is not None else grpc.aio.insecure_channel(host_and_port)
        )
        self.stubBank = bank_query_grpc.QueryStub(grpc_client)
        self.auth_client = AuthGrpcClient(query_grpc_client)
        self.txs = TxGrpcClient(grpc_client)
        self.stubCosmosTendermint = tendermint_query_grpc.ServiceStub(
            grpc_client
        )

    @staticmethod
    def calculate_quantums(
            size: float,
            atomic_resolution: int,
            step_base_quantums: int,
    ):
        raw_quantums = size * 10 ** (-1 * atomic_resolution)
        return int(max(raw_quantums, step_base_quantums))

    @staticmethod
    def calculate_subticks(
            price: float,
            atomic_resolution: int,
            quantum_conversion_exponent: int,
            subticks_per_tick: int
    ):
        exponent = atomic_resolution - quantum_conversion_exponent - CONSTANTS.QUOTE_QUANTUMS_ATOMIC_RESOLUTION
        raw_subticks = price * 10 ** (exponent)
        return int(max(raw_subticks, subticks_per_tick))

    def calculate_good_til_block_time(self, good_til_time_in_seconds: int) -> int:
        now = datetime.now()
        interval = timedelta(seconds=good_til_time_in_seconds)
        future = now + interval
        return int(future.timestamp())

    def get_sequence(self):
        current_seq = self.sequence
        self.sequence += 1
        return current_seq

    def get_number(self):
        return self.number

    async def trading_account_sequence(self) -> int:
        if not self._is_trading_account_initialized:
            await self.initialize_trading_account()
        return self.get_sequence()

    async def trading_account_number(self) -> int:
        if not self._is_trading_account_initialized:
            await self.initialize_trading_account()
        return self.get_number()

    async def initialize_trading_account(self):
        await self.query_account()
        self._is_trading_account_initialized = True

    def generate_good_til_fields(
            self,
            order_flags: int,
            good_til_block: int,
            good_til_time_in_seconds: int,
    ) -> Tuple[int, int]:
        if order_flags == CONSTANTS.ORDER_FLAGS_LONG_TERM:
            return 0, self.calculate_good_til_block_time(good_til_time_in_seconds)
        else:
            return good_til_block, 0

    async def latest_block(self) -> tendermint_query.GetLatestBlockResponse:
        '''
        Get lastest block

        :returns: Response, containing block information

        '''
        return await self.stubCosmosTendermint.GetLatestBlock(
            tendermint_query.GetLatestBlockRequest()
        )

    async def send_message(
            self,
            msg: _message.Message,
    ):
        tx = Transaction()
        tx.add_message(msg)
        return await self.prepare_and_broadcast_basic_transaction(
            tx=tx,
            memo=None,
        )

    async def cancel_order(
            self,
            client_id: int,
            clob_pair_id: int,
            order_flags: int,
            good_til_block_time: int,
    ):

        subaccount_id = SubaccountId(owner=self._dydx_v4_chain_address, number=self._subaccount_num)
        order_id = OrderId(
            subaccount_id=subaccount_id,
            client_id=client_id,
            order_flags=order_flags,
            clob_pair_id=int(clob_pair_id)
        )
        msg = MsgCancelOrder(
            order_id=order_id,
            good_til_block_time=good_til_block_time
        )
        result = await self.send_message(msg)
        return result

    async def place_order(
            self,
            market,
            type,
            side,
            price,
            size,
            client_id: int,
            post_only: bool,
            reduce_only: bool = False,
            good_til_time_in_seconds: int = 6000,
    ):

        clob_pair_id = self._connector._margin_fractions[market]["clob_pair_id"]
        atomic_resolution = self._connector._margin_fractions[market]["atomicResolution"]
        step_base_quantums = self._connector._margin_fractions[market]["stepBaseQuantums"]
        quantum_conversion_exponent = self._connector._margin_fractions[market]["quantumConversionExponent"]
        subticks_per_tick = self._connector._margin_fractions[market]["subticksPerTick"]

        order_side = Order.SIDE_BUY if side == "BUY" else Order.SIDE_SELL
        quantums = self.calculate_quantums(size, atomic_resolution, step_base_quantums)
        subticks = self.calculate_subticks(price, atomic_resolution, quantum_conversion_exponent, subticks_per_tick)
        order_flags = CONSTANTS.ORDER_FLAGS_SHORT_TERM if type == "MARKET" else CONSTANTS.ORDER_FLAGS_LONG_TERM

        if type == "MARKET":
            time_in_force = CONSTANTS.TIME_IN_FORCE_IOC
            latest_block_result = await self.latest_block()
            good_til_block = latest_block_result.block.header.height + 1 + 10
        else:
            good_til_block = 0
            if post_only:
                time_in_force = CONSTANTS.TIME_IN_FORCE_POST_ONLY
            else:
                time_in_force = CONSTANTS.TIME_IN_FORCE_UNSPECIFIED

        good_til_block, good_til_block_time = self.generate_good_til_fields(
            order_flags,
            good_til_block,
            good_til_time_in_seconds,
        )
        client_metadata = 1 if type == "MARKET" else 0
        condition_type = Order.CONDITION_TYPE_UNSPECIFIED
        conditional_order_trigger_subticks = 0

        subaccount_id = SubaccountId(owner=self._dydx_v4_chain_address, number=self._subaccount_num)

        order_id = OrderId(
            subaccount_id=subaccount_id,
            client_id=client_id,
            order_flags=order_flags,
            clob_pair_id=int(clob_pair_id)
        )
        order = Order(
            order_id=order_id,
            side=order_side,
            quantums=quantums,
            subticks=subticks,
            good_til_block=good_til_block,
            time_in_force=time_in_force,
            reduce_only=reduce_only,
            client_metadata=client_metadata,
            condition_type=condition_type,
            conditional_order_trigger_subticks=conditional_order_trigger_subticks,
        ) if (good_til_block != 0) else Order(
            order_id=order_id,
            side=order_side,
            quantums=quantums,
            subticks=subticks,
            good_til_block_time=good_til_block_time,
            time_in_force=time_in_force,
            reduce_only=reduce_only,
            client_metadata=client_metadata,
            condition_type=condition_type,
            conditional_order_trigger_subticks=conditional_order_trigger_subticks,
        )
        msg = MsgPlaceOrder(order=order)
        return await self.send_message(msg=msg)

    async def query_account(self):
        request = QueryAccountRequest(address=self._dydx_v4_chain_address)
        response = await self.auth_client.Account(request)

        account = BaseAccount()
        if not response.account.Is(BaseAccount.DESCRIPTOR):
            raise RuntimeError("Unexpected account type returned from query")
        response.account.Unpack(account)
        self.sequence = account.sequence
        self.number = account.account_number
        return account.sequence, account.account_number

    async def prepare_and_broadcast_basic_transaction(
            self,
            tx: "Transaction",  # type: ignore # noqa: F821
            memo: Optional[str] = None,
    ):
        async with self.transaction_lock:
            # query the account information for the sender
            sequence = await self.trading_account_sequence()
            number = await self.trading_account_number()
            # finally, build the final transaction that will be executed with the correct gas and fee values
            tx.seal(
                SigningCfg.direct(self._private_key, sequence),
                fee=f"{CONSTANTS.TX_FEE}{CONSTANTS.FEE_DENOMINATION}",
                gas_limit=CONSTANTS.TX_GAS_LIMIT,
                memo=memo,
            )
            tx.sign(self._private_key, CONSTANTS.CHAIN_ID, number)
            tx.complete()

            broadcast_req = BroadcastTxRequest(
                tx_bytes=tx.tx.SerializeToString(), mode=BroadcastMode.BROADCAST_MODE_SYNC
            )
            result = await self.send_tx_sync_mode(broadcast_req)
            err_msg = result.get("raw_log", "")
            if CONSTANTS.ACCOUNT_SEQUENCE_MISMATCH_ERROR in err_msg:
                await self.initialize_trading_account()

            return result

    async def send_tx_sync_mode(self, broadcast_req):
        resp = await self.txs.BroadcastTx(broadcast_req)
        result = json_format.MessageToDict(
            message=resp,
            always_print_fields_with_no_presence=True,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
        ).get("tx_response", {})
        return result



import asyncio
import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import Dict, List

import aiohttp

from hummingbot.core.utils.async_utils import safe_gather

PARROT_MINER_BASE_URL = "https://api.hummingbot.io/bounty/"

s_decimal_0 = Decimal("0")


@dataclass
class CampaignSummary:
    market_id: int = 0
    trading_pair: str = ""
    exchange_name: str = ""
    spread_max: Decimal = s_decimal_0
    payout_asset: str = ""
    liquidity: Decimal = s_decimal_0
    liquidity_usd: Decimal = s_decimal_0
    active_bots: int = 0
    reward_per_wk: Decimal = s_decimal_0
    apy: Decimal = s_decimal_0


def logger():
    return logging.getLogger(__name__)


async def get_campaign_summary(exchange: str, trading_pairs: List[str] = []) -> Dict[str, CampaignSummary]:
    results = {}
    try:
        campaigns = await get_active_campaigns(exchange, trading_pairs)
        tasks = [get_market_last_snapshot(m_id) for m_id in campaigns]
        snapshots = await safe_gather(*tasks, return_exceptions=True)
        for snapshot in snapshots:
            if isinstance(snapshot, Exception):
                raise snapshot
            if 'status' in snapshot and snapshot.get('status') != "success":
                logger().warning(
                    f"Snapshot info for {trading_pairs} is not available, please verify that this is a valid campaign pair for this exchange")
                continue
            if "market_snapshot" in snapshot:
                snapshot = snapshot.get("market_snapshot")
                market_id = int(snapshot.get("market_id"))
                campaign = campaigns[market_id]
                campaign.apy = Decimal(snapshot.get("annualized_return"))
                oov = snapshot.get("summary_stats").get("open_volume")
                campaign.liquidity = Decimal(oov.get("oov_eligible_ask")) + Decimal(oov.get("oov_eligible_bid"))
                campaign.liquidity_usd = campaign.liquidity * Decimal(oov.get("base_asset_usd_rate"))
                campaign.active_bots = int(oov.get("bots"))
        results = {c.trading_pair: c for c in campaigns.values()}
    except asyncio.CancelledError:
        raise
    except Exception:
        logger().error("Unexpected error while requesting data from Hummingbot API.", exc_info=True)
    return results


async def get_market_snapshots(market_id: int):
    async with aiohttp.ClientSession() as client:
        url = f"{PARROT_MINER_BASE_URL}charts/market_band?market_id={market_id}&chart_interval=1"
        resp = await client.get(url)
        resp_json = await resp.json()

    if not resp_json or "status" not in resp_json or resp_json.get("status") == "error":
        logger().warning("Could not get market snapshots from Hummingbot API"
                         f" (returned response '{resp_json}').")
        return None
    return resp_json


async def get_market_last_snapshot(market_id: int):
    data = await get_market_snapshots(market_id)
    data = sorted(list(set([d.get("timestamp") for d in data.get('data')])))

    await asyncio.sleep(0.5)

    async with aiohttp.ClientSession() as client:
        url = f"{PARROT_MINER_BASE_URL}user/single_snapshot?market_id={market_id}&timestamp={data[-1]}&aggregate_period=1m"
        resp = await client.get(url)
        resp_json = await resp.json()
    return resp_json


async def get_active_campaigns(exchange: str, trading_pairs: List[str] = []) -> Dict[int, CampaignSummary]:
    campaigns = {}
    async with aiohttp.ClientSession() as client:
        campaigns_url = f"{PARROT_MINER_BASE_URL}campaigns"
        resp = await client.get(campaigns_url)
        resp_json = await resp.json()

    if not resp_json or "status" not in resp_json or resp_json.get("status") == "error":
        logger().warning("Could not get active campaigns from Hummingbot API"
                         f" (returned response '{resp_json}').")
    else:
        for campaign_retval in resp_json["campaigns"]:
            for market in campaign_retval["markets"]:
                if not are_same_entity(exchange, market["exchange_name"]):
                    continue
                t_pair = f"{market['base_asset']}-{market['quote_asset']}"
                if trading_pairs and t_pair not in trading_pairs:
                    continue
                campaign = CampaignSummary()
                campaign.market_id = int(market["market_id"])
                campaign.trading_pair = t_pair
                campaign.exchange_name = exchange
                campaigns[campaign.market_id] = campaign

        campaigns = await get_active_markets(campaigns)

    return campaigns


async def get_active_markets(campaigns: Dict[int, CampaignSummary]) -> Dict[int, CampaignSummary]:
    async with aiohttp.ClientSession() as client:
        markets_url = f"{PARROT_MINER_BASE_URL}markets"
        resp = await client.get(markets_url)
        resp_json = await resp.json()

    if not resp_json or "status" not in resp_json or resp_json.get("status") == "error":
        logger().warning("Could not get active markets from Hummingbot API"
                         f" (returned response '{resp_json}').")
    else:
        for markets_retval in resp_json["markets"]:
            market_id = int(markets_retval["market_id"])
            if market_id in campaigns:
                campaigns[market_id].active_bots = markets_retval["bots"]
                for bounty_period in markets_retval["active_bounty_periods"]:
                    campaigns[market_id].reward_per_wk = Decimal(str(bounty_period["budget"]["bid"])) + Decimal(
                        str(bounty_period["budget"]["ask"]))
                    campaigns[market_id].spread_max = Decimal(str(bounty_period["spread_max"])) / Decimal("100")
                    campaigns[market_id].payout_asset = bounty_period["payout_asset"]

    return campaigns


def are_same_entity(hb_exchange_name: str, parrot_exchange_name: str) -> bool:
    return hb_exchange_name.replace("_", "") == parrot_exchange_name.replace("_", "")



from typing import Any, Callable, Dict, Optional

import hummingbot.connector.exchange.foxbit.foxbit_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import TimeSynchronizerRESTPreProcessor
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def public_rest_url(path_url: str,
                    domain: str = CONSTANTS.DEFAULT_DOMAIN,
                    ) -> str:
    """
    Creates a full URL for provided public REST endpoint
    :param path_url: a public REST endpoint
    :param domain: The default value is "com.br". Not in use at this time.
    :return: the full URL to the endpoint
    """
    return f"{CONSTANTS.REST_URL}/rest/{CONSTANTS.PUBLIC_API_VERSION}/{path_url}"


def public_rest_v2_url(path_url: str) -> str:
    """
    Creates a full URL for provided public REST V2 endpoint
    :param path_url: a public REST endpoint
    :return: the full URL to the endpoint
    """
    return f"{CONSTANTS.REST_V2_URL}/{path_url}"


def private_rest_url(path_url: str,
                     domain: str = CONSTANTS.DEFAULT_DOMAIN,
                     ) -> str:
    """
    Creates a full URL for provided private REST endpoint
    :param path_url: a private REST endpoint
    :param domain: The default value is "com.br". Not in use at this time.
    :return: the full URL to the endpoint
    """
    return f"{CONSTANTS.REST_URL}/rest/{CONSTANTS.PRIVATE_API_VERSION}/{path_url}"


def rest_endpoint_url(full_url: str,
                      ) -> str:
    """
    Creates a REST endpoint
    :param full_url: a full url
    :return: the URL endpoint
    """
    url_size = len(CONSTANTS.REST_URL)
    return full_url[url_size:]


def websocket_url() -> str:
    """
    Creates a full URL for provided WebSocket endpoint
    :return: the full URL to the endpoint
    """
    return f"wss://{CONSTANTS.WSS_URL}/"


def format_ws_header(header: Dict[str, Any]) -> Dict[str, Any]:
    retValue = {}
    retValue.update(CONSTANTS.WS_HEADER.copy())
    retValue.update(header)
    return retValue


def build_api_factory(throttler: Optional[AsyncThrottler] = None,
                      time_synchronizer: Optional[TimeSynchronizer] = None,
                      domain: str = CONSTANTS.DEFAULT_DOMAIN,
                      time_provider: Optional[Callable] = None,
                      auth: Optional[AuthBase] = None,
                      ) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    time_synchronizer = time_synchronizer or TimeSynchronizer()
    time_provider = time_provider or (lambda: get_current_server_time(
        throttler=throttler,
        domain=domain,
    ))
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        rest_pre_processors=[
            TimeSynchronizerRESTPreProcessor(synchronizer=time_synchronizer, time_provider=time_provider),
        ])
    return api_factory


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    api_factory = WebAssistantsFactory(throttler=throttler)
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(throttler: Optional[AsyncThrottler] = None,
                                  domain: str = CONSTANTS.DEFAULT_DOMAIN,
                                  ) -> float:
    throttler = throttler or create_throttler()
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    rest_assistant = await api_factory.get_rest_assistant()
    response = await rest_assistant.execute_request(url=public_rest_url(path_url=CONSTANTS.SERVER_TIME_PATH_URL,
                                                                        domain=domain),
                                                    method=RESTMethod.GET,
                                                    throttler_limit_id=CONSTANTS.SERVER_TIME_PATH_URL,
                                                    )
    server_time = response["timestamp"]
    return server_time



"""
ConfigVar is variable that is configured by the user via the hummingbot client that controls the trading behavior
of the bot. The client provides a screen prompt to the user, then the user provides input. This input is validated
by ConfigVar.
"""

import inspect
from typing import Callable, Optional, Union

# function types passed into ConfigVar
RequiredIf = Callable[[str], Optional[bool]]
Validator = Callable[[str], Optional[str]]
Prompt = Union[Callable[[str], Optional[str]], Optional[str]]
OnValidated = Callable


class ConfigVar:
    def __init__(self,
                 key: str,
                 prompt: Prompt,
                 is_secure: bool = False,
                 default: any = None,
                 type_str: str = "str",
                 # Whether this config will be prompted during the setup process
                 required_if: RequiredIf = lambda: True,
                 validator: Validator = lambda *args: None,
                 on_validated: OnValidated = lambda *args: None,
                 # Whether to prompt a user for value when new strategy config file is created
                 prompt_on_new: bool = False,
                 # Whether this is a config var used in connect command
                 is_connect_key: bool = False,
                 printable_key: str = None):
        self.prompt = prompt
        self.key = key
        self.value = None
        self.is_secure = is_secure
        self.default = default
        self.type = type_str
        self._required_if = required_if
        self._validator = validator
        self._on_validated = on_validated
        self.prompt_on_new = prompt_on_new
        self.is_connect_key = is_connect_key
        self.printable_key = printable_key

    async def get_prompt(self):
        """
        Call self.prompt if it is a function, otherwise return it as a value.
        """
        if inspect.iscoroutinefunction(self.prompt):
            return await self.prompt()
        elif inspect.isfunction(self.prompt):
            return self.prompt()
        else:
            return self.prompt

    @property
    def required(self) -> bool:
        assert callable(self._required_if)
        return self._required_if()

    async def validate(self, value: str) -> Optional[str]:
        """
        Validate user input against the function self._validator, if it is valid, then call self._on_validated,
        if it is invalid, then return the error message.
        """
        assert callable(self._validator)
        assert callable(self._on_validated)
        if self.required and (value is None or value == ""):
            return "Value is required."
        err_msg = None
        if inspect.iscoroutinefunction(self._validator):
            err_msg = await self._validator(value)
        elif inspect.isfunction(self._validator):
            err_msg = self._validator(value)
        if err_msg is None and self._validator is not None:
            if inspect.iscoroutinefunction(self._on_validated):
                await self._on_validated(value)
            elif inspect.isfunction(self._on_validated):
                self._on_validated(value)
        return err_msg



import asyncio
import logging
from typing import Any, Dict, List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.hyperliquid_spot_candles import constants as CONSTANTS
from hummingbot.logger import HummingbotLogger


class HyperliquidPerpetualCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        self._tokens = None
        self._base = trading_pair.split("-")[0]
        # For HIP-3 markets, convert dex prefix to lowercase (e.g., "XYZ:XYZ100" -> "xyz:XYZ100")
        if ":" in self._base:
            deployer, coin = self._base.split(":")
            self._base_asset = f"{deployer.lower()}:{coin}"
        else:
            self._base_asset = self._base
        super().__init__(trading_pair, interval, max_records)
        self._ping_timeout = CONSTANTS.PING_TIMEOUT
        self._ping_task: Optional[asyncio.Task] = None

    @property
    def name(self):
        return f"hyperliquid_perpetual_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url

    @property
    def candles_url(self):
        return self.rest_url

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        self._tokens = await rest_assistant.execute_request(url=self.rest_url,
                                                            method=RESTMethod.POST,
                                                            throttler_limit_id=self.rest_url,
                                                            data=CONSTANTS.HEALTH_CHECK_PAYLOAD)
        return NetworkStatus.CONNECTED

    def _rest_payload(self, **kwargs):
        return {
            "type": "candleSnapshot",
            "req": {
                "interval": CONSTANTS.INTERVALS[self.interval],
                "coin": self._base_asset,
                "startTime": kwargs["start_time"] * 1000,
                "endTime": kwargs["end_time"] * 1000,
            }
        }

    @property
    def _rest_method(self):
        return RESTMethod.POST

    @property
    def _rest_throttler_limit_id(self):
        return self.rest_url

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return False

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return False

    def get_exchange_trading_pair(self, trading_pair):
        return trading_pair.replace("-", "")

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = None) -> dict:
        pass  # No need to implement this method for Hyperliquid

    def _get_rest_candles_headers(self):
        return {"Content-Type": "application/json"}

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        return [
            [self.ensure_timestamp_in_seconds(row["t"]), row["o"], row["h"], row["l"], row["c"], row["v"], 0.,
             row["n"], 0., 0.] for row in data
        ]

    def ws_subscription_payload(self):
        interval = CONSTANTS.INTERVALS[self.interval]
        payload = {
            "method": "subscribe",
            "subscription": {
                "type": "candle",
                "coin": self._base_asset,
                "interval": interval
            },
        }
        return payload

    def _parse_websocket_message(self, data):
        candles_row_dict: Dict[str, Any] = {}
        if data is not None and data.get("channel") == "candle":
            candle = data["data"]
            candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(candle["t"])
            candles_row_dict["open"] = candle["o"]
            candles_row_dict["low"] = candle["l"]
            candles_row_dict["high"] = candle["h"]
            candles_row_dict["close"] = candle["c"]
            candles_row_dict["volume"] = candle["v"]
            candles_row_dict["quote_asset_volume"] = 0.
            candles_row_dict["n_trades"] = candle["n"]
            candles_row_dict["taker_buy_base_volume"] = 0.
            candles_row_dict["taker_buy_quote_volume"] = 0.
            return candles_row_dict

    @property
    def _ping_payload(self):
        return CONSTANTS.PING_PAYLOAD

    async def _ping_loop(self, websocket_assistant: WSAssistant):
        """
        Sends ping messages at regular intervals to keep the WebSocket connection alive.
        Hyperliquid requires proactive pinging - the server will close the connection
        if it doesn't receive a ping within a certain time window.
        """
        try:
            while True:
                await asyncio.sleep(self._ping_timeout)
                ping_request = WSJSONRequest(payload=self._ping_payload)
                await websocket_assistant.send(request=ping_request)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().debug(f"Ping loop error: {e}")

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the candles events and starts the ping loop.
        """
        await super()._subscribe_channels(ws)
        # Start the ping loop to keep the connection alive
        if self._ping_task is not None:
            self._ping_task.cancel()
        self._ping_task = safe_ensure_future(self._ping_loop(ws))

    async def _on_order_stream_interruption(self, websocket_assistant: Optional[WSAssistant] = None):
        """
        Clean up the ping task when the WebSocket connection is interrupted.
        """
        if self._ping_task is not None:
            self._ping_task.cancel()
            self._ping_task = None
        await super()._on_order_stream_interruption(websocket_assistant)



# A single source of truth for constant variables related to the exchange
# https://api.btcmarkets.net/doc/v3#section/General-notes

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "btc_markets"
DEFAULT_DOMAIN = "btc_markets"
# Base URL
REST_URLS = {"btc_markets": "https://api.btcmarkets.net/"}

WSS_V1_PUBLIC_URL = {"btc_markets": "wss://socket.btcmarkets.net/v2"}

WSS_PRIVATE_URL = {"btc_markets": "wss://socket.btcmarkets.net/v2"}

REST_API_VERSION = "v3"
WS_PING_TIMEOUT = 10

HBOT_ORDER_ID_PREFIX = "BTCM-"
MAX_ORDER_ID_LEN = 32
HBOT_BROKER_ID = "Hummingbot"

SIDE_BUY = "BUY"
SIDE_SELL = "SELL"

TIME_IN_FORCE_GTC = "GTC"

# REST API Public Endpoints
ACCOUNTS_URL = f"{REST_API_VERSION}/accounts"
BALANCE_URL = f"{ACCOUNTS_URL}/me/balances"
FEES_URL = f"{ACCOUNTS_URL}/me/trading-fees"
MARKETS_URL = f"{REST_API_VERSION}/markets"
ORDERS_URL = f"{REST_API_VERSION}/orders"
BATCH_ORDERS_URL = f"{REST_API_VERSION}/batchorders"
TRADES_URL = f"{REST_API_VERSION}/trades"
SERVER_TIME_PATH_URL = f"{REST_API_VERSION}/time"

WS_CONNECTION_LIMIT_ID = "WSConnection"
WS_REQUEST_LIMIT_ID = "WSRequest"
WS_SUBSCRIPTION_LIMIT_ID = "WSSubscription"
WS_LOGIN_LIMIT_ID = "WSLogin"

# Websocket event types
TICK = "tick"
DIFF_EVENT_TYPE = "orderbookUpdate"
SNAPSHOT_EVENT_TYPE = "orderbook"
ORDER_CHANGE_EVENT_TYPE = "orderChange"
TRADE_EVENT_TYPE = "trade"
FUND_CHANGE_EVENT_TYPE = "fundChange"
HEARTBEAT = "heartbeat"
ERROR = "error"
SUBSCRIBE = "subscribe"

# Order States
ORDER_STATE = {
    "Accepted": OrderState.APPROVED,
    "Placed": OrderState.OPEN,
    "Partially Matched": OrderState.PARTIALLY_FILLED,
    "Fully Matched": OrderState.FILLED,
    "Partially Cancelled": OrderState.PENDING_CANCEL,
    "Cancelled": OrderState.CANCELED,
    "Failed": OrderState.FAILED,
}

WS_HEARTBEAT_TIME_INTERVAL = 30

RATE_LIMITS = [
    RateLimit(WS_CONNECTION_LIMIT_ID, limit=3, time_interval=10),
    RateLimit(WS_REQUEST_LIMIT_ID, limit=3, time_interval=10),
    RateLimit(WS_SUBSCRIPTION_LIMIT_ID, limit=3, time_interval=10),
    RateLimit(WS_LOGIN_LIMIT_ID, limit=1, time_interval=15),
    RateLimit(limit_id=ACCOUNTS_URL, limit=50, time_interval=10),
    RateLimit(limit_id=BALANCE_URL, limit=50, time_interval=10),
    RateLimit(limit_id=FEES_URL, limit=50, time_interval=10),
    RateLimit(limit_id=MARKETS_URL, limit=150, time_interval=10),
    RateLimit(limit_id=ORDERS_URL, limit=50, time_interval=10),
    RateLimit(limit_id=BATCH_ORDERS_URL, limit=50, time_interval=10),
    RateLimit(limit_id=TRADES_URL, limit=50, time_interval=10),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=50, time_interval=10)
]
"""
Rate Limits - https://api.btcmarkets.net/doc/v3#section/General-Notes
Rate Limits ws - https://docs.btcmarkets.net/v3/#tag/WS_Overview
"""

# Error codes
INVALID_TIME_WINDOW = "InvalidTimeWindow"
INVALID_TIMESTAMP = "InvalidTimestamp"
INVALID_AUTH_TIMESTAMP = "InvalidAuthTimestamp"
INVALID_AUTH_SIGNATURE = "InvalidAuthSignature"
ORDER_NOT_FOUND = "OrderNotFound"
INVALID_ORDERID = "InvalidOrderId"



"""
Exceptions used in the Hummingbot codebase.
"""


class HummingbotBaseException(Exception):
    """
    Most errors raised in Hummingbot should inherit this class so we can
    differentiate them from errors that come from dependencies.
    """


class ArgumentParserError(HummingbotBaseException):
    """
    Unable to parse a command (like start, stop, etc) from the hummingbot client
    """


class OracleRateUnavailable(HummingbotBaseException):
    """
    Asset value from third party is unavailable
    """


class InvalidScriptModule(HummingbotBaseException):
    """
    The file does not contain a ScriptBase subclass
    """


class InvalidController(HummingbotBaseException):
    """
    The file does not contain a ControllerBase subclass
    """



import logging
from typing import Any, Dict, List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.data_feed.candles_feed.aevo_perpetual_candles import constants as CONSTANTS
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.logger import HummingbotLogger


class AevoPerpetualCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)
        self._ping_timeout = CONSTANTS.PING_TIMEOUT
        self._current_ws_candle: Optional[Dict[str, Any]] = None

    async def initialize_exchange_data(self):
        if self._ex_trading_pair is None:
            self._ex_trading_pair = self.get_exchange_trading_pair(self._trading_pair)

    @property
    def name(self):
        return f"aevo_perpetual_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        base_asset = trading_pair.split("-")[0]
        return f"{base_asset}-PERP"

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = None) -> dict:
        if limit is None:
            limit = self.candles_max_result_per_rest_request
            if start_time is not None and end_time is not None:
                expected_records = int((end_time - start_time) / self.interval_in_seconds) + 1
                limit = min(limit, expected_records)

        params = {
            "instrument_name": self._ex_trading_pair,
            "resolution": CONSTANTS.INTERVALS[self.interval],
            "limit": limit,
        }
        if start_time is not None:
            params["start_timestamp"] = int(start_time * 1e9)
        if end_time is not None:
            params["end_timestamp"] = int(end_time * 1e9)
        return params

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        history = []
        if data is not None:
            history = data.get("history", [])
        if history:
            candles = []
            for timestamp, price in reversed(history):
                candle_price = float(price)
                candles.append([
                    self.ensure_timestamp_in_seconds(timestamp),
                    candle_price,
                    candle_price,
                    candle_price,
                    candle_price,
                    0.,
                    0.,
                    0.,
                    0.,
                    0.,
                ])
            return candles
        return []

    def ws_subscription_payload(self):
        return {
            "op": "subscribe",
            "data": [f"{CONSTANTS.WS_TICKER_CHANNEL}:{self._ex_trading_pair}"],
        }

    def _parse_websocket_message(self, data):
        if data is None:
            return None
        channel = data.get("channel")
        if channel != f"{CONSTANTS.WS_TICKER_CHANNEL}:{self._ex_trading_pair}":
            return None
        tickers = data.get("data", {}).get("tickers", [])
        if not tickers:
            return None
        ticker = tickers[0]
        price = None
        mark = ticker.get("mark") or {}
        if "price" in mark:
            price = mark["price"]
        elif "index_price" in ticker:
            price = ticker["index_price"]
        if price is None:
            return None
        timestamp = data.get("data", {}).get("timestamp") or data.get("write_ts")
        if timestamp is None:
            return None
        timestamp_s = self.ensure_timestamp_in_seconds(timestamp)
        candle_timestamp = int(timestamp_s - (timestamp_s % self.interval_in_seconds))
        candle_price = float(price)

        if self._current_ws_candle is None or candle_timestamp > self._current_ws_candle["timestamp"]:
            self._current_ws_candle = {
                "timestamp": candle_timestamp,
                "open": candle_price,
                "high": candle_price,
                "low": candle_price,
                "close": candle_price,
                "volume": 0.,
                "quote_asset_volume": 0.,
                "n_trades": 0.,
                "taker_buy_base_volume": 0.,
                "taker_buy_quote_volume": 0.,
            }
        elif candle_timestamp == self._current_ws_candle["timestamp"]:
            self._current_ws_candle["high"] = max(self._current_ws_candle["high"], candle_price)
            self._current_ws_candle["low"] = min(self._current_ws_candle["low"], candle_price)
            self._current_ws_candle["close"] = candle_price
        else:
            return None

        return self._current_ws_candle

    @property
    def _ping_payload(self):
        return CONSTANTS.PING_PAYLOAD



from typing import Any, Dict

import numpy
from sqlalchemy import BigInteger, Column, Index, Integer, Text
from sqlalchemy.orm import relationship

from hummingbot.model import HummingbotBase
from hummingbot.model.decimal_type_decorator import SqliteDecimal


class Order(HummingbotBase):
    __tablename__ = "Order"
    __table_args__ = (Index("o_config_timestamp_index",
                            "config_file_path", "creation_timestamp"),
                      Index("o_market_trading_pair_timestamp_index",
                            "market", "symbol", "creation_timestamp"),
                      Index("o_market_base_asset_timestamp_index",
                            "market", "base_asset", "creation_timestamp"),
                      Index("o_market_quote_asset_timestamp_index",
                            "market", "quote_asset", "creation_timestamp"))

    id = Column(Text, primary_key=True, nullable=False)
    config_file_path = Column(Text, nullable=False)
    strategy = Column(Text, nullable=False)
    market = Column(Text, nullable=False)
    symbol = Column(Text, nullable=False)
    base_asset = Column(Text, nullable=False)
    quote_asset = Column(Text, nullable=False)
    creation_timestamp = Column(BigInteger, nullable=False)
    order_type = Column(Text, nullable=False)
    amount = Column(SqliteDecimal(6), nullable=False)
    leverage = Column(Integer, nullable=False, default=1)
    price = Column(SqliteDecimal(6), nullable=False)
    last_status = Column(Text, nullable=False)
    last_update_timestamp = Column(BigInteger, nullable=False)
    exchange_order_id = Column(Text, nullable=True)
    position = Column(Text, nullable=True)
    status = relationship("OrderStatus", back_populates="order")
    trade_fills = relationship("TradeFill", back_populates="order")

    def __repr__(self) -> str:
        return f"Order(id={self.id}, config_file_path='{self.config_file_path}', strategy='{self.strategy}', " \
               f"market='{self.market}', symbol='{self.symbol}', base_asset='{self.base_asset}', " \
               f"quote_asset='{self.quote_asset}', creation_timestamp={self.creation_timestamp}, " \
               f"order_type='{self.order_type}', amount={self.amount}, leverage={self.leverage}, " \
               f"price={self.price}, last_status='{self.last_status}', " \
               f"last_update_timestamp={self.last_update_timestamp}), " \
               f"exchange_order_id={self.exchange_order_id}, position={self.position}"

    @staticmethod
    def to_bounty_api_json(order: "Order") -> Dict[str, Any]:
        return {
            "order_id": order.id,
            "price": numpy.format_float_positional(order.price),
            "quantity": numpy.format_float_positional(order.amount),
            "symbol": order.symbol,
            "market": order.market,
            "order_timestamp": order.creation_timestamp,
            "order_type": order.order_type,
            "base_asset": order.base_asset,
            "quote_asset": order.quote_asset,
            "raw_json": {
            }
        }



"""
XRPL Worker Pool Module

This module provides concurrent worker pools for XRPL operations:
- XRPLQueryWorkerPool: Concurrent read-only queries
- XRPLVerificationWorkerPool: Concurrent transaction verification
- XRPLTransactionWorkerPool: Concurrent preparation, serialized submission via pipeline

Architecture:
- Each pool manages multiple worker coroutines
- Workers acquire clients from the node pool per task (round-robin)
- Clients are released back to the pool after task completion
- Transaction submissions are serialized through a shared pipeline to prevent sequence conflicts

Error Handling:
- On client error: try reconnect same client
- If reconnect fails: get new healthy client from pool
- If no healthy client available: wait with timeout
- If timeout expires: fail the task with error
"""

import asyncio
import logging
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, TypeVar

from xrpl.asyncio.clients import AsyncWebsocketClient
from xrpl.asyncio.clients.exceptions import XRPLWebsocketException
from xrpl.asyncio.transaction import XRPLReliableSubmissionException, sign
from xrpl.core.binarycodec import encode
from xrpl.models import Request, Response, SubmitOnly, Transaction, Tx
from xrpl.wallet import Wallet

from hummingbot.connector.exchange.xrpl import xrpl_constants as CONSTANTS
from hummingbot.connector.exchange.xrpl.xrpl_utils import (
    XRPLConnectionError,
    XRPLNodePool,
    XRPLTimeoutError,
    _wait_for_final_transaction_outcome,
    autofill,
)
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.xrpl.xrpl_transaction_pipeline import XRPLTransactionPipeline


# Type variable for generic pool result type
T = TypeVar("T")


# ============================================
# Result Dataclasses
# ============================================

@dataclass
class TransactionSubmitResult:
    """Result of a transaction submission."""
    success: bool
    signed_tx: Optional[Transaction] = None
    response: Optional[Response] = None
    prelim_result: Optional[str] = None
    exchange_order_id: Optional[str] = None
    error: Optional[str] = None
    tx_hash: Optional[str] = None

    @property
    def is_queued(self) -> bool:
        """Check if transaction was queued (terQUEUED)."""
        return self.prelim_result == "terQUEUED"

    @property
    def is_accepted(self) -> bool:
        """Check if transaction was accepted (tesSUCCESS or terQUEUED)."""
        return self.prelim_result in ("tesSUCCESS", "terQUEUED")


@dataclass
class TransactionVerifyResult:
    """Result of a transaction verification."""
    verified: bool
    response: Optional[Response] = None
    final_result: Optional[str] = None
    error: Optional[str] = None


@dataclass
class QueryResult:
    """Result of a query operation."""
    success: bool
    response: Optional[Response] = None
    error: Optional[str] = None


# ============================================
# Worker Task Dataclass
# ============================================

@dataclass
class WorkerTask(Generic[T]):
    """Represents a task submitted to a worker pool."""
    task_id: str
    request: Any
    future: asyncio.Future
    created_at: float = field(default_factory=time.time)
    timeout: float = CONSTANTS.WORKER_TASK_TIMEOUT
    max_queue_time: float = CONSTANTS.WORKER_MAX_QUEUE_TIME

    @property
    def is_expired(self) -> bool:
        """Check if the task has waited too long in the queue.

        Note: This only checks queue wait time, not processing time.
        Processing timeout is handled separately in the worker loop.
        """
        return (time.time() - self.created_at) > self.max_queue_time


# ============================================
# Pool Statistics
# ============================================

@dataclass
class PoolStats:
    """Statistics for a worker pool."""
    pool_name: str
    num_workers: int
    tasks_completed: int = 0
    tasks_failed: int = 0
    tasks_pending: int = 0
    total_latency_ms: float = 0.0
    client_reconnects: int = 0
    client_failures: int = 0

    @property
    def avg_latency_ms(self) -> float:
        """Calculate average task latency."""
        total = self.tasks_completed + self.tasks_failed
        if total == 0:
            return 0.0
        return self.total_latency_ms / total

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for logging/monitoring."""
        return {
            "pool_name": self.pool_name,
            "num_workers": self.num_workers,
            "tasks_completed": self.tasks_completed,
            "tasks_failed": self.tasks_failed,
            "tasks_pending": self.tasks_pending,
            "avg_latency_ms": round(self.avg_latency_ms, 2),
            "client_reconnects": self.client_reconnects,
            "client_failures": self.client_failures,
        }


# ============================================
# Base Worker Pool Class
# ============================================

class XRPLWorkerPoolBase(ABC, Generic[T]):
    """
    Abstract base class for XRPL worker pools.

    Features:
    - Multiple concurrent worker coroutines
    - Task queue management
    - Round-robin client acquisition per task
    - Error handling with reconnect/retry logic
    - Statistics tracking

    Subclasses must implement:
    - _process_task(): Execute the actual work for a task
    """
    _logger: Optional[HummingbotLogger] = None

    def __init__(
        self,
        node_pool: XRPLNodePool,
        pool_name: str,
        num_workers: int,
        max_queue_size: int = CONSTANTS.WORKER_POOL_TASK_QUEUE_SIZE,
    ):
        """
        Initialize the worker pool.

        Args:
            node_pool: The XRPL node pool for getting connections
            pool_name: Name of the pool for logging
            num_workers: Number of concurrent worker coroutines
            max_queue_size: Maximum pending tasks in the queue
        """
        self._node_pool = node_pool
        self._pool_name = pool_name
        self._num_workers = num_workers
        self._max_queue_size = max_queue_size

        # Task queue
        self._task_queue: asyncio.Queue[WorkerTask] = asyncio.Queue(maxsize=max_queue_size)

        # Worker tasks
        self._worker_tasks: List[asyncio.Task] = []
        self._running = False
        self._started = False  # Track if pool was ever started (for lazy init)

        # Statistics
        self._stats = PoolStats(pool_name=pool_name, num_workers=num_workers)

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    @property
    def is_running(self) -> bool:
        """Check if the pool is running."""
        return self._running

    @property
    def stats(self) -> PoolStats:
        """Get pool statistics."""
        self._stats.tasks_pending = self._task_queue.qsize()
        return self._stats

    async def start(self):
        """Start the worker pool."""
        if self._running:
            self.logger().warning(f"[{self._pool_name}] Pool is already running")
            return

        self._running = True
        self._started = True

        # Create worker tasks
        for i in range(self._num_workers):
            worker_task = asyncio.create_task(self._worker_loop(worker_id=i))
            self._worker_tasks.append(worker_task)

        self.logger().debug(
            f"[{self._pool_name}] Started pool with {self._num_workers} workers"
        )

    async def stop(self):
        """Stop the worker pool and cancel pending tasks."""
        if not self._running:
            return

        self._running = False
        self.logger().debug(f"[{self._pool_name}] Stopping pool...")

        # Cancel all worker tasks
        for worker_task in self._worker_tasks:
            worker_task.cancel()

        # Wait for workers to finish
        if self._worker_tasks:
            await asyncio.gather(*self._worker_tasks, return_exceptions=True)
        self._worker_tasks.clear()

        # Cancel pending tasks in queue
        cancelled_count = 0
        while not self._task_queue.empty():
            try:
                task = self._task_queue.get_nowait()
                if not task.future.done():
                    task.future.cancel()
                    cancelled_count += 1
            except asyncio.QueueEmpty:
                break

        self.logger().debug(
            f"[{self._pool_name}] Pool stopped, cancelled {cancelled_count} pending tasks"
        )

    async def _ensure_started(self):
        """Ensure the pool is started (lazy initialization)."""
        if not self._started:
            await self.start()

    async def submit(self, request: Any, timeout: Optional[float] = None) -> T:
        """
        Submit a task to the worker pool.

        Args:
            request: The request to process
            timeout: Optional timeout override

        Returns:
            The result of processing the task

        Raises:
            asyncio.QueueFull: If the task queue is full
            Exception: Any exception from task processing
        """
        # Lazy start
        await self._ensure_started()

        task_id = str(uuid.uuid4())[:8]
        future: asyncio.Future = asyncio.get_event_loop().create_future()
        task = WorkerTask(
            task_id=task_id,
            request=request,
            future=future,
            timeout=timeout or CONSTANTS.WORKER_TASK_TIMEOUT,
        )

        try:
            self._task_queue.put_nowait(task)
            self.logger().debug(
                f"[{self._pool_name}] Task {task_id} queued "
                f"(queue_size={self._task_queue.qsize()}, request_type={type(request).__name__})"
            )
        except asyncio.QueueFull:
            self.logger().error(
                f"[{self._pool_name}] Task queue full, rejecting task {task_id}"
            )
            raise

        # Wait for result - no timeout here since queue wait time should not count
        # The processing timeout is applied in the worker loop when the task is picked up
        try:
            result = await future
            return result
        except asyncio.CancelledError:
            self.logger().warning(f"[{self._pool_name}] Task {task_id} was cancelled")
            raise
        except asyncio.TimeoutError:
            # This timeout comes from the worker loop during processing
            self.logger().error(f"[{self._pool_name}] Task {task_id} timed out during processing")
            raise

    async def _worker_loop(self, worker_id: int):
        """
        Worker loop that processes tasks from the queue.

        Args:
            worker_id: Identifier for this worker
        """
        self.logger().debug(f"[{self._pool_name}] Worker {worker_id} started and ready")

        while self._running:
            try:
                # Get next task with timeout
                try:
                    task = await asyncio.wait_for(
                        self._task_queue.get(),
                        timeout=1.0
                    )
                except asyncio.TimeoutError:
                    continue

                # Skip expired or cancelled tasks
                if task.future.done():
                    self.logger().debug(
                        f"[{self._pool_name}] Worker {worker_id} skipping cancelled task {task.task_id}"
                    )
                    continue

                if task.is_expired:
                    queue_time = time.time() - task.created_at
                    self.logger().warning(
                        f"[{self._pool_name}] Worker {worker_id} skipping expired task {task.task_id} "
                        f"(waited {queue_time:.1f}s in queue, max={task.max_queue_time}s)"
                    )
                    if not task.future.done():
                        task.future.set_exception(
                            asyncio.TimeoutError(f"Task {task.task_id} expired after {queue_time:.1f}s in queue")
                        )
                    self._stats.tasks_failed += 1
                    continue

                # Process the task with timeout (timeout only applies to processing, not queue wait)
                queue_time = time.time() - task.created_at
                self.logger().debug(
                    f"[{self._pool_name}] Worker {worker_id} processing task {task.task_id} "
                    f"(queued for {queue_time:.1f}s)"
                )
                start_time = time.time()
                try:
                    # Apply timeout only to the actual processing
                    result = await asyncio.wait_for(
                        self._process_task_with_retry(task, worker_id),
                        timeout=task.timeout
                    )
                    elapsed_ms = (time.time() - start_time) * 1000

                    if not task.future.done():
                        task.future.set_result(result)

                    self._stats.tasks_completed += 1
                    self._stats.total_latency_ms += elapsed_ms

                    self.logger().debug(
                        f"[{self._pool_name}] Worker {worker_id} completed task {task.task_id} "
                        f"in {elapsed_ms:.1f}ms (success={getattr(result, 'success', True)})"
                    )

                except asyncio.TimeoutError:
                    elapsed_ms = (time.time() - start_time) * 1000
                    self.logger().error(
                        f"[{self._pool_name}] Worker {worker_id} task {task.task_id} timed out "
                        f"after {elapsed_ms:.1f}ms processing (timeout={task.timeout}s)"
                    )
                    if not task.future.done():
                        task.future.set_exception(
                            asyncio.TimeoutError(
                                f"Task {task.task_id} timed out after {elapsed_ms:.1f}ms processing"
                            )
                        )
                    self._stats.tasks_failed += 1

                except Exception as e:
                    elapsed_ms = (time.time() - start_time) * 1000

                    if not task.future.done():
                        task.future.set_exception(e)

                    self._stats.tasks_failed += 1
                    self._stats.total_latency_ms += elapsed_ms

                    self.logger().error(
                        f"[{self._pool_name}] Worker {worker_id} failed task {task.task_id} "
                        f"after {elapsed_ms:.1f}ms: {e}"
                    )

            except asyncio.CancelledError:
                break
            except Exception as e:
                self.logger().error(
                    f"[{self._pool_name}] Worker {worker_id} unexpected error: {e}"
                )

        self.logger().debug(f"[{self._pool_name}] Worker {worker_id} stopped")

    async def _process_task_with_retry(self, task: WorkerTask, worker_id: int) -> T:
        """
        Process a task with client error handling and retry.

        Error handling flow:
        1. Get client from pool
        2. Try to process task
        3. On error: try reconnect same client
        4. If reconnect fails: get new healthy client
        5. If no healthy client: wait with timeout
        6. If timeout: raise error

        Args:
            task: The task to process
            worker_id: The worker processing the task

        Returns:
            The result of processing
        """
        client = None
        reconnect_attempts = 0
        max_reconnect = CONSTANTS.WORKER_CLIENT_RECONNECT_ATTEMPTS

        while True:
            try:
                # Get client if we don't have one
                if client is None:
                    client = await self._get_client_with_timeout(worker_id)

                # Process the task
                return await self._process_task(task, client)

            except (XRPLConnectionError, XRPLWebsocketException) as e:
                self.logger().warning(
                    f"[{self._pool_name}] Worker {worker_id} connection error: {e}"
                )

                # Try to reconnect
                if reconnect_attempts < max_reconnect:
                    reconnect_attempts += 1
                    self._stats.client_reconnects += 1

                    self.logger().debug(
                        f"[{self._pool_name}] Worker {worker_id} attempting reconnect "
                        f"({reconnect_attempts}/{max_reconnect})"
                    )

                    try:
                        # Try to reconnect the existing client
                        if client is not None:
                            await client.open()
                            self.logger().debug(
                                f"[{self._pool_name}] Worker {worker_id} reconnected successfully"
                            )
                            continue
                    except Exception as reconnect_error:
                        self.logger().warning(
                            f"[{self._pool_name}] Worker {worker_id} reconnect failed: {reconnect_error}"
                        )

                    # Get a new client
                    client = None
                    continue

                # Max reconnects reached, fail
                self._stats.client_failures += 1
                raise XRPLConnectionError(
                    f"Failed after {max_reconnect} reconnect attempts: {e}"
                )

            except Exception:
                # Non-connection error, don't retry
                raise

    async def _get_client_with_timeout(self, worker_id: int) -> AsyncWebsocketClient:
        """
        Get a healthy client from the node pool with timeout.

        Args:
            worker_id: The worker requesting the client

        Returns:
            A healthy WebSocket client

        Raises:
            XRPLConnectionError: If no client available within timeout
        """
        timeout = CONSTANTS.WORKER_CLIENT_RETRY_TIMEOUT
        start_time = time.time()

        while (time.time() - start_time) < timeout:
            try:
                client = await self._node_pool.get_client(use_burst=False)
                return client
            except Exception as e:
                self.logger().warning(
                    f"[{self._pool_name}] Worker {worker_id} failed to get client: {e}"
                )
                await asyncio.sleep(0.5)

        self._stats.client_failures += 1
        raise XRPLConnectionError(
            f"No healthy client available after {timeout}s timeout"
        )

    @abstractmethod
    async def _process_task(self, task: WorkerTask, client: AsyncWebsocketClient) -> T:
        """
        Process a single task. Must be implemented by subclasses.

        Args:
            task: The task to process
            client: The client to use for the operation

        Returns:
            The result of the task
        """
        pass


# ============================================
# Query Worker Pool
# ============================================

class XRPLQueryWorkerPool(XRPLWorkerPoolBase[QueryResult]):
    """
    Worker pool for concurrent read-only XRPL queries.

    Use for: AccountInfo, AccountTx, AccountObjects, Tx, ServerInfo, etc.
    """

    def __init__(
        self,
        node_pool: XRPLNodePool,
        num_workers: int = CONSTANTS.QUERY_WORKER_POOL_SIZE,
    ):
        super().__init__(
            node_pool=node_pool,
            pool_name="QueryPool",
            num_workers=num_workers,
        )

    async def _process_task(
        self,
        task: WorkerTask,
        client: AsyncWebsocketClient,
    ) -> QueryResult:
        """
        Execute a query against the XRPL.

        Args:
            task: The task containing the request
            client: The client to use

        Returns:
            QueryResult with success status and response
        """
        request: Request = task.request
        request_type = type(request).__name__

        try:
            response = await client._request_impl(request)

            if response.is_successful():
                return QueryResult(success=True, response=response)
            else:
                error = response.result.get("error", "Unknown error")
                error_message = response.result.get("error_message", "")
                full_error = f"{error}: {error_message}" if error_message else error
                self.logger().warning(
                    f"[QueryPool] {request_type} request returned error: {full_error}"
                )
                return QueryResult(success=False, response=response, error=full_error)

        except XRPLConnectionError:
            # Re-raise connection errors for retry handling
            raise
        except KeyError as e:
            # KeyError can occur if the connection reconnects during the request,
            # which clears _open_requests in the XRPL library
            self.logger().warning(f"[QueryPool] Request lost during client reconnection: {e}")
            raise XRPLConnectionError(f"Request lost during reconnection: {e}")
        except Exception as e:
            # Provide more context in error messages
            error_msg = f"{request_type} query failed: {type(e).__name__}: {str(e)}"
            self.logger().error(f"[QueryPool] {error_msg}")
            return QueryResult(success=False, error=error_msg)


# ============================================
# Verification Worker Pool
# ============================================

class XRPLVerificationWorkerPool(XRPLWorkerPoolBase[TransactionVerifyResult]):
    """
    Worker pool for concurrent transaction verification.

    Verifies that transactions have been finalized on the ledger.
    """

    def __init__(
        self,
        node_pool: XRPLNodePool,
        num_workers: int = CONSTANTS.VERIFICATION_WORKER_POOL_SIZE,
    ):
        super().__init__(
            node_pool=node_pool,
            pool_name="VerifyPool",
            num_workers=num_workers,
        )

    async def submit_verification(
        self,
        signed_tx: Transaction,
        prelim_result: str,
        timeout: float = CONSTANTS.VERIFY_TX_TIMEOUT,
    ) -> TransactionVerifyResult:
        """
        Submit a transaction for verification.

        Args:
            signed_tx: The signed transaction to verify
            prelim_result: The preliminary result from submission
            timeout: Maximum time to wait for verification

        Returns:
            TransactionVerifyResult with verification outcome
        """
        # Package the verification request
        request = {
            "signed_tx": signed_tx,
            "prelim_result": prelim_result,
        }
        return await self.submit(request, timeout=timeout)

    async def _process_task(
        self,
        task: WorkerTask,
        client: AsyncWebsocketClient,
    ) -> TransactionVerifyResult:
        """
        Verify a transaction's finality on the ledger.

        Args:
            task: The task containing verification request
            client: The client to use

        Returns:
            TransactionVerifyResult with verification outcome
        """
        request = task.request
        signed_tx: Transaction = request["signed_tx"]
        prelim_result: str = request["prelim_result"]

        # Only verify transactions that have a chance of success
        if prelim_result not in ("tesSUCCESS", "terQUEUED"):
            self.logger().warning(
                f"[VerifyPool] Transaction prelim_result={prelim_result} indicates failure"
            )
            return TransactionVerifyResult(
                verified=False,
                error=f"Preliminary result {prelim_result} indicates failure",
            )

        tx_hash = signed_tx.get_hash()
        self.logger().debug(
            f"[VerifyPool] Starting verification for tx_hash={tx_hash[:16]}..."
        )

        try:
            # Try primary verification method
            result = await self._verify_with_wait(signed_tx, prelim_result, client, task.timeout)
            if result.verified:
                return result

            # Fallback to direct hash query
            self.logger().warning(
                f"[VerifyPool] Primary verification failed for {tx_hash[:16]}, "
                f"trying fallback query..."
            )
            return await self._verify_with_hash_query(tx_hash, client)

        except (XRPLConnectionError, XRPLWebsocketException):
            # Re-raise connection errors for retry handling at the worker level
            raise

        except Exception as e:
            self.logger().error(f"[VerifyPool] Verification error: {e}")
            return TransactionVerifyResult(
                verified=False,
                error=str(e),
            )

    async def _verify_with_wait(
        self,
        signed_tx: Transaction,
        prelim_result: str,
        client: AsyncWebsocketClient,
        timeout: float,
    ) -> TransactionVerifyResult:
        """Verify using the wait_for_final_transaction_outcome method."""
        try:
            response = await asyncio.wait_for(
                _wait_for_final_transaction_outcome(
                    transaction_hash=signed_tx.get_hash(),
                    client=client,
                    prelim_result=prelim_result,
                    last_ledger_sequence=signed_tx.last_ledger_sequence,
                ),
                timeout=timeout,
            )

            final_result = response.result.get("meta", {}).get("TransactionResult", "unknown")

            self.logger().debug(
                f"[VerifyPool] Transaction verified: "
                f"hash={signed_tx.get_hash()[:16]}, result={final_result}"
            )

            return TransactionVerifyResult(
                verified=True,
                response=response,
                final_result=final_result,
            )

        except XRPLReliableSubmissionException as e:
            self.logger().error(f"[VerifyPool] Transaction failed on-chain: {e}")
            return TransactionVerifyResult(
                verified=False,
                error=f"Transaction failed: {e}",
            )

        except asyncio.TimeoutError:
            self.logger().warning("[VerifyPool] Verification timed out")
            return TransactionVerifyResult(
                verified=False,
                error="Verification timed out",
            )

        except (XRPLConnectionError, XRPLWebsocketException):
            # Re-raise connection errors for retry handling
            raise

        except Exception as e:
            self.logger().warning(f"[VerifyPool] Verification error: {e}")
            return TransactionVerifyResult(
                verified=False,
                error=str(e),
            )

    async def _verify_with_hash_query(
        self,
        tx_hash: str,
        client: AsyncWebsocketClient,
        max_attempts: int = 5,
        poll_interval: float = 3.0,
    ) -> TransactionVerifyResult:
        """Fallback verification by querying transaction hash directly."""
        self.logger().debug(
            f"[VerifyPool] Fallback query for tx_hash={tx_hash[:16]}..."
        )

        for attempt in range(max_attempts):
            try:
                tx_request = Tx(transaction=tx_hash)
                response = await client._request_impl(tx_request)

                if not response.is_successful():
                    error = response.result.get("error", "unknown")
                    if error == "txnNotFound":
                        self.logger().debug(
                            f"[VerifyPool] tx_hash={tx_hash[:16]} not found, "
                            f"attempt {attempt + 1}/{max_attempts}"
                        )
                    else:
                        self.logger().warning(
                            f"[VerifyPool] Error querying tx_hash={tx_hash[:16]}: {error}"
                        )
                else:
                    result = response.result
                    if result.get("validated", False):
                        final_result = result.get("meta", {}).get("TransactionResult", "unknown")
                        self.logger().debug(
                            f"[VerifyPool] Transaction found and validated: "
                            f"tx_hash={tx_hash[:16]}, result={final_result}"
                        )
                        return TransactionVerifyResult(
                            verified=final_result == "tesSUCCESS",
                            response=response,
                            final_result=final_result,
                        )
                    else:
                        self.logger().debug(
                            f"[VerifyPool] tx_hash={tx_hash[:16]} found but not validated yet"
                        )

            except XRPLConnectionError:
                # Re-raise for retry handling
                raise
            except KeyError as e:
                # KeyError can occur if the connection reconnects during the request,
                # which clears _open_requests in the XRPL library
                self.logger().warning(f"[VerifyPool] Request lost during client reconnection: {e}")
                raise XRPLConnectionError(f"Request lost during reconnection: {e}")
            except XRPLWebsocketException:
                # Re-raise for retry handling - websocket is not open
                raise
            except Exception as e:
                self.logger().warning(
                    f"[VerifyPool] Exception querying tx_hash={tx_hash[:16]}: {e}"
                )

            # Wait before next attempt
            if attempt < max_attempts - 1:
                await asyncio.sleep(poll_interval)

        return TransactionVerifyResult(
            verified=False,
            error=f"Transaction not found after {max_attempts} attempts",
        )


# ============================================
# Transaction Worker Pool
# ============================================

class XRPLTransactionWorkerPool(XRPLWorkerPoolBase[TransactionSubmitResult]):
    """
    Worker pool for transaction submissions.

    Features:
    - Concurrent transaction preparation (autofill, signing)
    - Serialized submission through a shared pipeline
    - Handles sequence error retries

    The pipeline ensures only one transaction is submitted at a time,
    preventing sequence number race conditions.
    """

    def __init__(
        self,
        node_pool: XRPLNodePool,
        wallet: Wallet,
        pipeline: "XRPLTransactionPipeline",
        num_workers: int = CONSTANTS.TX_WORKER_POOL_SIZE,
    ):
        """
        Initialize the transaction worker pool.

        Args:
            node_pool: The XRPL node pool
            wallet: The wallet for signing transactions
            pipeline: The shared transaction pipeline
            num_workers: Number of concurrent workers
        """
        super().__init__(
            node_pool=node_pool,
            pool_name=f"TxPool[{wallet.classic_address[:8]}]",
            num_workers=num_workers,
        )
        self._wallet = wallet
        self._pipeline = pipeline

    async def submit_transaction(
        self,
        transaction: Transaction,
        fail_hard: bool = True,
        max_retries: int = CONSTANTS.PLACE_ORDER_MAX_RETRY,
    ) -> TransactionSubmitResult:
        """
        Submit a transaction through the pool.

        Args:
            transaction: The unsigned transaction to submit
            fail_hard: Whether to use fail_hard mode
            max_retries: Maximum retry attempts for sequence errors

        Returns:
            TransactionSubmitResult with submission outcome
        """
        request = {
            "transaction": transaction,
            "fail_hard": fail_hard,
            "max_retries": max_retries,
        }
        return await self.submit(request, timeout=CONSTANTS.SUBMIT_TX_TIMEOUT * max_retries)

    async def _process_task(
        self,
        task: WorkerTask,
        client: AsyncWebsocketClient,
    ) -> TransactionSubmitResult:
        """
        Process a transaction submission task.

        This handles retries for sequence errors but delegates
        the actual submission to the pipeline for serialization.

        Args:
            task: The task containing transaction details
            client: The client to use (for autofill)

        Returns:
            TransactionSubmitResult with submission outcome
        """
        request = task.request
        transaction: Transaction = request["transaction"]
        fail_hard: bool = request.get("fail_hard", True)
        max_retries: int = request.get("max_retries", CONSTANTS.PLACE_ORDER_MAX_RETRY)

        submission_id = task.task_id
        self.logger().debug(f"[{self._pool_name}] Starting submission {submission_id}")

        submit_retry = 0
        last_error = None

        while submit_retry < max_retries:
            try:
                # Submit through pipeline - this serializes all submissions
                result = await self._submit_through_pipeline(
                    transaction, fail_hard, submission_id, client
                )

                # Handle successful submission
                if result.is_accepted:
                    self.logger().debug(
                        f"[{self._pool_name}] Submission {submission_id} accepted: "
                        f"prelim_result={result.prelim_result}, tx_hash={result.tx_hash}"
                    )
                    return result

                # Handle sequence errors - retry with fresh autofill
                if result.prelim_result in CONSTANTS.SEQUENCE_ERRORS:
                    submit_retry += 1
                    retry_interval = (
                        CONSTANTS.PRE_SEQ_RETRY_INTERVAL
                        if result.prelim_result == "terPRE_SEQ"
                        else CONSTANTS.PLACE_ORDER_RETRY_INTERVAL
                    )
                    self.logger().debug(
                        f"[{self._pool_name}] {submission_id} got {result.prelim_result}. "
                        f"Waiting {retry_interval}s and retrying... "
                        f"(Attempt {submit_retry}/{max_retries})"
                    )
                    await asyncio.sleep(retry_interval)
                    continue

                # Handle transient errors - retry
                if result.prelim_result in CONSTANTS.TRANSIENT_RETRY_ERRORS:
                    submit_retry += 1
                    self.logger().debug(
                        f"[{self._pool_name}] {submission_id} got {result.prelim_result}. "
                        f"Retrying... (Attempt {submit_retry}/{max_retries})"
                    )
                    await asyncio.sleep(CONSTANTS.PLACE_ORDER_RETRY_INTERVAL)
                    continue

                # Other error - don't retry
                self.logger().error(
                    f"[{self._pool_name}] {submission_id} failed: prelim_result={result.prelim_result}"
                )
                return result

            except XRPLTimeoutError as e:
                # Timeout - DO NOT retry as transaction may have succeeded
                self.logger().error(
                    f"[{self._pool_name}] {submission_id} timed out: {e}. "
                    f"NOT retrying to avoid duplicate transactions."
                )
                return TransactionSubmitResult(
                    success=False,
                    error=f"Timeout: {e}",
                )

            except XRPLConnectionError:
                # Re-raise for retry handling at pool level
                raise

            except Exception as e:
                last_error = str(e)
                self.logger().error(f"[{self._pool_name}] {submission_id} error: {e}")
                submit_retry += 1
                if submit_retry < max_retries:
                    await asyncio.sleep(CONSTANTS.PLACE_ORDER_RETRY_INTERVAL)

        return TransactionSubmitResult(
            success=False,
            error=f"Max retries ({max_retries}) reached: {last_error}",
        )

    async def _submit_through_pipeline(
        self,
        transaction: Transaction,
        fail_hard: bool,
        submission_id: str,
        client: AsyncWebsocketClient,
    ) -> TransactionSubmitResult:
        """
        Execute the actual submission through the pipeline.

        This ensures only one transaction is autofilled/submitted at a time,
        preventing sequence number race conditions.

        Args:
            transaction: The transaction to submit
            fail_hard: Whether to use fail_hard mode
            submission_id: Identifier for tracing
            client: The client to use

        Returns:
            TransactionSubmitResult with outcome
        """
        async def _do_submit():
            self.logger().debug(f"[{self._pool_name}] {submission_id}: Autofilling transaction...")
            filled_tx = await autofill(transaction, client)

            self.logger().debug(
                f"[{self._pool_name}] {submission_id}: Autofill done, "
                f"sequence={filled_tx.sequence}, "
                f"last_ledger={filled_tx.last_ledger_sequence}"
            )

            # Sign the transaction
            signed_tx = sign(filled_tx, self._wallet)
            tx_hash = signed_tx.get_hash()

            self.logger().debug(
                f"[{self._pool_name}] {submission_id}: Submitting to XRPL, tx_hash={tx_hash[:8]}..."
            )

            # Submit
            tx_blob = encode(signed_tx.to_xrpl())
            response = await client._request_impl(
                SubmitOnly(tx_blob=tx_blob, fail_hard=fail_hard),
                timeout=CONSTANTS.REQUEST_TIMEOUT,
            )

            return signed_tx, response

        # Submit through the pipeline
        signed_tx, response = await self._pipeline.submit(_do_submit(), submission_id)

        prelim_result = response.result.get("engine_result", "UNKNOWN")
        tx_hash = signed_tx.get_hash()
        tx_hash_prefix = tx_hash[:6]
        exchange_order_id = f"{signed_tx.sequence}-{signed_tx.last_ledger_sequence}-{tx_hash_prefix}"

        self.logger().debug(
            f"[{self._pool_name}] {submission_id}: Complete, "
            f"exchange_order_id={exchange_order_id}, prelim_result={prelim_result}"
        )

        return TransactionSubmitResult(
            success=prelim_result in ("tesSUCCESS", "terQUEUED"),
            signed_tx=signed_tx,
            response=response,
            prelim_result=prelim_result,
            exchange_order_id=exchange_order_id,
            tx_hash=tx_hash,
        )



from decimal import Decimal

from hummingbot.connector.constants import SECOND
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

DEFAULT_DOMAIN = "aevo_perpetual"
BROKER_ID = "HBOT"

MAX_ORDER_ID_LEN = 32

PERPETUAL_INSTRUMENT_TYPE = "PERPETUAL"

MARKET_ORDER_SLIPPAGE = Decimal("0.01")

# REST endpoints
BASE_URL = "https://api.aevo.xyz"
TESTNET_BASE_URL = "https://api-testnet.aevo.xyz"

# WS endpoints
WSS_URL = "wss://ws.aevo.xyz"
TESTNET_WSS_URL = "wss://ws-testnet.aevo.xyz"

# Public REST API
PING_PATH_URL = "/time"
MARKETS_PATH_URL = "/markets"
ORDERBOOK_PATH_URL = "/orderbook"
FUNDING_PATH_URL = "/funding"
INSTRUMENT_PATH_URL = "/instrument"

# Private REST API
ACCOUNT_PATH_URL = "/account"
PORTFOLIO_PATH_URL = "/portfolio"
POSITIONS_PATH_URL = "/positions"
ORDERS_PATH_URL = "/orders"
ORDER_PATH_URL = "/orders/{order_id}"
ORDERS_ALL_PATH_URL = "/orders-all"
TRADE_HISTORY_PATH_URL = "/trade-history"
ACCOUNT_LEVERAGE_PATH_URL = "/account/leverage"
ACCOUNT_ACCUMULATED_FUNDINGS_PATH_URL = "/account/accumulated-fundings"

# WS channels
WS_ORDERBOOK_CHANNEL = "orderbook-100ms"
WS_TRADE_CHANNEL = "trades"
WS_TICKER_CHANNEL = "ticker-500ms"
WS_BOOK_TICKER_CHANNEL = "book-ticker"
WS_INDEX_CHANNEL = "index"

WS_ORDERS_CHANNEL = "orders"
WS_FILLS_CHANNEL = "fills"
WS_POSITIONS_CHANNEL = "positions"

WS_HEARTBEAT_TIME_INTERVAL = 30

NOT_EXIST_ERROR = "ORDER_DOES_NOT_EXIST"
REDUCE_ONLY_REJECTION_ERRORS = {
    "NO_POSITION_REDUCE_ONLY",
    "ORDER_EXCEEDS_CAPACITY_OF_REDUCE_ONLY",
    "INVALID_DIRECTION_REDUCE_ONLY",
}

# Order states
ORDER_STATE = {
    "opened": OrderState.OPEN,
    "partial": OrderState.PARTIALLY_FILLED,
    "filled": OrderState.FILLED,
    "cancelled": OrderState.CANCELED,
}

RATE_LIMITS = [
    RateLimit(limit_id=PING_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=MARKETS_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=ORDERBOOK_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=FUNDING_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=INSTRUMENT_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=ACCOUNT_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=PORTFOLIO_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=POSITIONS_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=ORDERS_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=ORDERS_ALL_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=TRADE_HISTORY_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=ACCOUNT_LEVERAGE_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=ACCOUNT_ACCUMULATED_FUNDINGS_PATH_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=WSS_URL, limit=10, time_interval=SECOND),
    RateLimit(limit_id=TESTNET_WSS_URL, limit=10, time_interval=SECOND),
]
